import pandas as pd
import json
import os
from loguru import logger
from uuid import uuid4

from .utils.ut_autprog import AutProgUtils
from .utils.ut import Utils
from .utils.ut_auth import AuthData


class Programming():

    def __init__(self, endpoint: str, client: object, auth_data: AuthData) -> None:

        self.raiseException = client.raiseException
        self.defaults = client.defaults

        self.auth_data = auth_data
        self.endpoint = endpoint
        self.proxies = client.proxies

        return

    def getVersion(self):
        """
        Returns name and version of the responsible micro service
        """

        return Utils._getServiceVersion(self, 'programming')

    def functions(self) -> pd.DataFrame:
        """ Get available programming service functions"""

        graphQLString = f'''query functions {{
            functions {{
                name
                functionId
                languageVersion
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None:
            return

        df = pd.json_normalize(result['functions'])
        return df

    def createFunction(
        self,
        name: str,
        languageVersion: str,
        description: str = None,
        files: list = None,
        deploy: bool = False,
        deploymentName: str = None,
        envVars: dict = None,
        memoryLimit: str = '250Mi',
        cpuLimit: float = 0.2,
        secrets: list = None
    ) -> str:
        """ 
        Creates a function with possibility to commit files and deployment
        in one step. The function Id is returned.

        Parameters:
        ----------
        name : str
            The name of the function, which is also taken for deployment, if the option is 
            chosen. The deployment function name is converted to small letters. Special characters 
            will be removed.
        languageVersion : str
            Choose between a languange and its version, e.g. 'PYTHON_3_9' and 'CSHARP_NET_6_0'.
        description : str
            Additional description to the function.
        files : list
            A list of full file paths to be committed.
        deploy : bool
            If True, the function will be deployed. The files argument must not be 
            None in this case.
        deploymentName:
            The name of the deployed function. If left to default (None) the function name 
            will be used. Use small letters only and no special characters.
        envVars : dict
            A Dictionary of Environment variables provided to the deployment function. All values
            will be converted to string.
        secrets: list = None
            A list of kubernetes secrets to be provided to the deployment function. 
            The secrets have to exist in the function's namespace.
            Secrets of type 'Opaque' are mounted automatically inside the function container at
            /var/openfaas/secrets/<key> as file(s) with the value as content.

        Examples of a function for Python:
        ---------
        >>> createFunction('myfunction', 'PYTHON_3_9')
        >>> files = [path1, path2, path3]
            vars = {'var1': 42, 'var2': 'full_scale'}
            createFunction('myfunction', 'PYTHON_3_9', files=files, deploy=True, 
                envVars=vars)
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            graphQLString = f'''mutation createFunction {{
                createFunction(input:{{
                    name: "{name}"
                    languageVersion: {languageVersion}
                    description: "{description}"
                    }}) {{
                    functionId
                    errors {{
                        message
                        code
                        }}
                    }}
                }}
            '''

            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result['createFunction']['errors']:
                Utils._listGraphQlErrors(
                    result, 'createInventory', correlationId)
                return

            functionId = result['createFunction']['functionId']

            if files == None:
                logger.info(
                    f"Function with id {functionId} created. No files committed")
                return functionId
            else:
                if type(files) != list:
                    msg = "Files must be of type list!"
                    if self.raiseException:
                        raise Exception(msg)
                    else:
                        logger.error(msg)
                        return

                self.commitFunctionFiles(functionId, files)
                logger.info(f"Function with id {functionId} created.")
                if deploy == False:
                    return functionId
                else:
                    if deploymentName == None:
                        deploymentName = name.lower()
                    self.deployFunction(
                        functionId, deploymentName, envVars, memoryLimit, cpuLimit, secrets)
                    return functionId

    def commitFunctionFiles(
        self,
        functionId: str,
        files: list = None,
        deploy: bool = False,
        deploymentName: str = None,
        envVars: dict = None,
        memoryLimit: str = '250Mi',
        cpuLimit: float = 0.2,
        secrets: list = None
    ) -> None:
        """
        Upload programming files to an existing function. 

        Parameters:
        ----------
        functionId : str
            Id of an existing function.
        files : list
            A list of full file paths to be committed.
        deploy : bool
            If True, the function will be deployed after committing the files. 
        deploymentName:
            The name of the deployed function. Use small letters only and no special characters.
        envVars : dict
            A dictionary of environment variables provided to the deployed function. All values
            will be converted to string.
        memoryLimit: str = '250Mi'
            The memory that is assigned for execution of this function. 
        cpuLimit: float = 0.2
            The assigned amount of CPU capacity to this function: note, that a too high capacity
            might affect other services negatively.
        secrets: list = None
            A list of kubernetes secrets to be provided to the deployment function. 
            The secrets have to exist in the function's namespace.
            Secrets of type 'Opaque' are mounted automatically inside the function container at
            /var/openfaas/secrets/<key> as file(s) with the value as content.


        Example:
        --------
        >>> files = [path1, path2, path3]
            vars = {'var1': 42, 'var2': 'full_scale'}
            commitFunctionFiles('3dba276e3c8645838c2d598043cab057', files=files, deploy=True, 
                deploymentName='myfunction', envVars=vars)
        """
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            fileList = AutProgUtils._upsetFiles(files)
            if fileList == None:
                return

            graphQLString = f'''mutation commitFunctionFiles {{
                commitFunctionFiles(input:{{
                    functionId: "{functionId}"
                    upsetFiles: [
                        {fileList}
                    ]

                    }}) {{
                    errors {{
                        message
                        code
                        }}
                    }}
                }}
            '''

            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return
            else:
                logger.info(
                    f"Committed files: {AutProgUtils._getFileNames(files)}")
                if deploy == False:
                    return functionId
                else:
                    if deploymentName == None:
                        msg = f"No deployment name was provided for deployment. Function id: {functionId}"
                        if self.raiseException:
                            raise Exception(msg)
                        else:
                            logger.error(msg)
                            return
                    self.deployFunction(
                        functionId, deploymentName, envVars, memoryLimit, cpuLimit, secrets)

            return result

    def deployFunction(
            self,
            functionId: str,
            deploymentName: str,
            envVars: dict = None,
            memoryLimit: str = '250Mi',
            cpuLimit: float = 0.2,
            secrets: list = None):
        """
        Deploys a function to make it executable.

        Parameters:
        ----------
        functionId: str
            Id of an existing function.
        deploymentName:
            The name of the deployed function. Use small letters only and no special characters.
        envVars: dict = None
            A Dictionary of Environment variables provided to the deployment function. 
        memoryLimit: str = '250M'
            The memory that is assigned for execution of this function.   
        cpuLimit: loat = 0.2
            The assigned amount of CPU capacity to this function: note, that a too high capacity
            might affect other services negatively.
        secrets: list = None
            A list of kubernetes secrets to be provided to the deployment function. 
            The secrets have to exist in the function's namespace.
            Secrets of type 'Opaque' are mounted automatically inside the function container at
            /var/openfaas/secrets/<key> as file(s) with the value as content.

        Example:
        --------
        >>> vars = {'exec_timeout': '2m', 'read_timeout': '2m','write_timeout': '2m'}
            deployFunction('3dba276e3c8645838c2d598043cab057', deploymentName=myfunction, 
            automationTopic='autFunction', envVars=vars)
        """
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            if envVars == None:
                _vars = ''
            else:
                _vars = AutProgUtils._varsToString(envVars, 'env')

            if secrets == None:
                _secrets = ''
            else:
                _secretsStrings = [f"\"{s}\"" for s in secrets]
                _secrets = f"secrets: [{','.join(_secretsStrings)}]"

            graphQLString = f'''mutation deployFunction {{
                deployFunction(
                    input: {{
                    functionId: "{functionId}"
                    functionName: "{deploymentName}"
                    memoryLimit: "{memoryLimit}"
                    cpuLimit: "{str(cpuLimit)}"
                    {_vars}
                    {_secrets}
                    }}
                ) {{
                    deploymentId
                    errors {{
                        code
                        message
                        }}
                    }}
                }}
                '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return
            else:
                deploymentId = result['deployFunction']['deploymentId']
                logger.info(
                    f"Deployed function with deploymentId {deploymentId}")
                return deploymentId

    def functionFiles(self, functionId: str, downloadPath: str = None) -> pd.DataFrame:
        """
        Shows files of a function.

        Parameters:
        ----------
        functionId : str
            Id of an existing function.

        Example:
        --------
        >>> functionFiles('3dba276e3c8645838c2d598043cab057')
        """

        if downloadPath == None:
            _contentBase64 = ''
        else:
            _contentBase64 = 'contentBase64'

        graphQLString = f''' query functionFiles {{
            functionFiles (functionId: "{functionId}") {{
                version
                files {{
                    fullname
                    {_contentBase64}
                    }}
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None:
            return

        df = pd.json_normalize(result['functionFiles'], meta=[
                               'version'], record_path=['files'])

        if downloadPath != None:
            if os.path.exists(downloadPath) == False:
                msg = f"Download path '{downloadPath}' does not exist."
                if self.raiseException:
                    raise Exception(msg)
                else:
                    logger.error(msg)
                    return

            else:
                fileList = []
                for file in result['functionFiles']['files']:
                    AutProgUtils._downloadFunctionFile(
                        downloadPath, file['fullname'], file['contentBase64'])
                    fileList.append(file['fullname'])
                logger.info(f"Downloaded {fileList} to {downloadPath}")
                del df['contentBase64']

        return df

    def executeFunction(self, deploymentId: str, inputVariables: dict = None) -> str:
        """
        Executes a function and returns its execution id

        Parameters:
        ----------
        deploymentId : str
            Id of a deployment. Can be retrieved by Programming.deployments().

        Example:
        --------
        >>> executeFunction('c877cc1b568a4c489aacdb4538b3f544')
        """
        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id=correlationId)

        if inputVariables == None:
            _vars = ''
        else:
            _vars = Utils._toGraphQL(json.dumps(inputVariables))

        graphQLString = f''' mutation executeFunction {{
            executeFunction(
                input: {{ 
                    deploymentId: {Utils._toGraphQL(deploymentId)},
                    input: {_vars}
                    }}
            ) {{
                executionId
                result {{
                    output
                    errorMessage
                    hasError
                }}
                errors {{
                    code
                    message
                    }}
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString, correlationId)
        context_logger.info("Executed function with result {}", result)

        if result == None:
            return

        return result

    def deleteFunction(self, functionId: str, force: bool = False) -> None:
        """
        Deletes a function.

        Parameters:
        ----------
        functionId : str
            Id of the function to be deleted.
        force : bool
            Use True to ignore confirmation.

        Example:
        --------
        >>> executeFunction('3dba276e3c8645838c2d598043cab057')
        """
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            if force == False:
                confirm = input(
                    f"Press 'y' to delete  function with id {functionId}")

            graphQLString = f''' mutation deleteFunction {{
                deleteFunction (input: {{
                    functionId: "{functionId}"
                }}) {{
                    errors {{
                        message
                    }}
                }}
            }}
            '''
            if force == True:
                confirm = 'y'
            if confirm == 'y':
                result = Utils._executeGraphQL(
                    self, graphQLString, correlationId)
                logger.info("Deleted function with result {}", result)
            else:
                return
            if result == None:
                return

            return result

    def deployments(self, functionId: str, fields: list = []) -> pd.DataFrame:
        """
        Shows deployments of a function as a DataFrame.
        Parameters:
        ----------
        functionId : str
            Id of the function.
        fields: list = []
            A list of fields to be queried. If None, all properties but 'log' will be queried.
            Available fields are: 'functionAggregateId', 'functionAggregateVersion', 'deploymentId', 'functionName', 'log', 'state'
        Example:
        --------
        >>> deployments('77725d50-78b5-4a16-992b-7f68f2dd261f')
        >>> deployments(
                '77725d50-78b5-4a16-992b-7f68f2dd261f',
                fields=['deploymentId', 'functionName', 'state']
                )
        >>> df = deployments('77725d50-78b5-4a16-992b-7f68f2dd261f', fields=['log'])
            logs = df['log'].values[0]
        """
        allFields = ['functionAggregateId', 'functionAggregateVersion', 'deploymentId', 'functionName', 'log', 'state']
        if not fields:
            fields = allFields.copy()
            fields.remove('log')
        else:
            for field in fields:
                if field not in allFields:
                    msg = f"Field '{field}' is not available. Choose from {allFields}"
                    if self.raiseException:
                        raise Exception(msg)
                    else:
                        logger.error(msg)
                        return

        graphQLString = f'''query deployments($functionId: UUID!) {{
            deployments(functionId: $functionId) {{
                    {' '.join(fields)}
                }}
            }}
            '''
        
        result = Utils._executeGraphQL(self, graphQLString, params={"functionId": functionId})
        if result == None:
            return

        df = pd.json_normalize(result['deployments'])
        return df



from re import match
from datetime import datetime
import pytz
import pandas as pd

class TimeUtils():

    def _inputTimestamp(timestamp, timeZone):

        if type(timestamp) == datetime:
            timestamp = pytz.timezone(timeZone).localize(timestamp).isoformat()
        elif type(timestamp) == pd._libs.tslibs.timestamps.Timestamp:
            timestamp = pytz.timezone(timeZone).localize(timestamp).isoformat()
        elif type(timestamp) == str:
            dateFormat = TimeUtils._dateFormat(timestamp)
            if dateFormat == None:
                raise Exception(f"Unsupported date format in timestamp: '{timestamp}'")
            if "%z" in dateFormat:
                timestamp = datetime.strptime(timestamp, dateFormat)
                timestamp = timestamp.astimezone(pytz.timezone(timeZone)).isoformat()
            else:
                timestamp = datetime.strptime(timestamp, dateFormat)
                timestamp = pytz.timezone(timeZone).localize(timestamp).isoformat()
                #if we need to specify dst time, while not having an offset, then we need another way to determine DST!
        return timestamp

    def _dateFormat(timestamp: str) -> str:

        # German
        if match(r'\d{2}\.\d{2}\.\d{4} \d{2}:\d{2}:\d{2}.\d{1,6} [-+]\d{2}:\d{2}', timestamp):
            return '%d.%m.%Y %H:%M:%S.%f %z'
        if match(r'\d{2}\.\d{2}\.\d{4} \d{2}:\d{2}:\d{2} [-+]\d{2}:\d{2}', timestamp):
            return '%d.%m.%Y %H:%M:%S %z'
        if match(r'\d{2}\.\d{2}\.\d{4} \d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%d.%m.%Y %H:%M:%S.%f'
        if match(r'\d{2}\.\d{2}\.\d{4} \d{2}:\d{2}:\d{2}', timestamp):
            return '%d.%m.%Y %H:%M:%S'
        if match(r'\d{2}\.\d{2}\.\d{4} \d{2}:\d{2}', timestamp):
            return '%d.%m.%Y %H:%M'
        if match(r'\d{2}\.\d{2}\.\d{4}', timestamp):
            return '%d.%m.%Y'

        # ISO
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{1,6}[-+]\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M:%S.%f%z'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}[-+]\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M:%S%z'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{1,6}Z', timestamp):
            return '%Y-%m-%dT%H:%M:%S.%f%z'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%Y-%m-%dT%H:%M:%S.%f'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z', timestamp):
            return '%Y-%m-%dT%H:%M:%S%z'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M:%S'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M'

        # English I
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{1,6} [-+]\d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M:%S.%f %z'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} [-+]\d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M:%S %z'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%Y-%m-%d %H:%M:%S.%f'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M:%S'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M'
        if match(r'\d{4}-\d{2}-\d{2}', timestamp):
            return '%Y-%m-%d'

        # English II
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2}.\d{1,6} [-+]\d{2}:\d{2}', timestamp):
            return '%Y/%m/%d %H:%M:%S.%f %z'
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} [-+]\d{2}:\d{2}', timestamp):
            return '%Y/%m/%d %H:%M:%S %z'
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%Y/%m/%d %H:%M:%S.%f'
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2}', timestamp):
            return '%Y/%m/%d %H:%M:%S'
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}', timestamp):
            return '%Y/%m/%d %H:%M'
        if match(r'\d{4}/\d{2}/\d{2}', timestamp):
            return '%Y/%m/%d'

        # English III
        if match(r'\d{1,2}/\d{1,2}/\d{4} \d{2}:\d{2}:\d{2} [-+]\d{2}:\d{2}', timestamp):
            return '%m/%d/%Y %H:%M:%S %z'
        if match(r'\d{1,2}/\d{1,2}/\d{4} \d{2}:\d{2} [-+]\d{2}:\d{2}', timestamp):
            return '%m/%d/%Y %H:%M %z'
        if match(r'\d{1,2}/\d{1,2}/\d{4} \d{2}:\d{2}:\d{2}', timestamp):
            return '%m/%d/%Y %H:%M:%S'
        if match(r'\d{1,2}/\d{1,2}/\d{4} \d{2}:\d{2}', timestamp):
            return '%m/%d/%Y %H:%M'
        if match(r'\d{1,2}/\d{1,2}/\d{4}', timestamp):
            return '%m/%d/%Y'

        return None



import json
import re
import base64
from uuid import uuid4
from pathlib import Path
from loguru import logger
from typing import Union

from requests_oauthlib import OAuth2Session

from .utils.ut import Utils
from .utils.ut_auth import AuthData


class Email:

    def __init__(self, endpoint: str, core: object, auth_data: AuthData) -> None:
        global client
        client = core

        self.raiseException = client.raiseException
        self.defaults = core.defaults
        self.structure = core.structure
        self.scheme = core.scheme
        self.endpoint = endpoint
        self.auth_data = auth_data

        self.header = {
            'AcceptEncoding': 'deflate',
            'Accept': 'application/json',
            'Content-type': 'application/json',
        }

        return

    def sendMail(
        self,
        to: Union[str, list],
        cc: Union[str, list] = None,
        bcc: Union[str, list] = None,
        subject: str = None,
        textBody: str = None,
        htmlBody: str = None,
        attachments: Union[str, list] = None
    ) -> None:
        """
        Sends an email via the TechStack email service.

        Parameters:
        ----------
        to: str|list
            One or more recipient addresses.
        cc: str|list = None
            One or more recipient addresses in CC.
        bcc: str|list = None
            One or more recipient addresses in BCC.
        subject: str = None
            The email subject.
        textBody: str = None
            The text body of the email.
        htmlBody: str = None
            A HTML body of the email. Not, if both textBody and htmlBody are used,
            only the htmlBody will be sent.
        attachments: str|list = None
            Provide one or more file paths to attach files to the email.

        Examples:
        >>> sendMail('gustav@mail.com', cc=['annette@mail.com', carl@mail.com], subject='Hello', textBody=text)
        >>> sendMail('gustav@mail.com', attachments=['report.pdf', 'data.xlsx']
        """

        correlationId = str(uuid4())

        if isinstance(to, str):
            to = [to]
        if isinstance(cc, str):
            cc = [cc]
        if isinstance(bcc, str):
            bcc = [bcc]
        if isinstance(attachments, str):
            attachments = [attachments]

        address_regex = re.compile(
            r'([A-Za-z0-9]+[.-_])*[A-Za-z0-9]+@[A-Za-z0-9-]+(\.[A-Z|a-z]{2,})+')
        for group in [to, cc, bcc]:
            if not group:
                continue
            for address in group:
                if not re.fullmatch(address_regex, address):
                    Utils._error(
                        self, f"Invalid email address '{address}'.", correlationId)
                    return

        _attachments = []
        if attachments:
            for filepath in attachments:
                if not Path(filepath).exists():
                    Utils._error(
                        self, f"File path '{filepath}' is not correct.", correlationId)
                    return
                with open(Path(filepath), 'rb') as file:
                    content = base64.b64encode(file.read()).decode('utf-8')
                    _attachments.append(
                        {'filename': Path(filepath).name, 'content': content})

        data = {
            'to': to,
            'subject': subject,
            'textBody': textBody,
            'htmlBody': htmlBody,
            'cc': cc,
            'bcc': bcc,
            'attachments': _attachments
        }

        data = json.dumps(data)

        with logger.contextualize(correlation_id=correlationId):
            def token_updater(new_token):
                self.auth_data.token = new_token

            email = OAuth2Session(self.auth_data.client_id, token=self.auth_data.token, 
                      auto_refresh_url=self.auth_data.token_url,
                      auto_refresh_kwargs={'client_id': self.auth_data.client_id},
                      token_updater=token_updater)
            
            response = email.post(
                url=self.endpoint, headers=self.header, data=data)
            if response.status_code >= 400:
                logger.info(
                    f'Email service not available. Response status {response.status}.')
            if response.status_code == 200:
                logger.info(f'Email sent to {to}.')

        return



from uuid import uuid4
import pandas as pd
from loguru import logger
import json

from .utils.ut import Utils
from .utils.ut_auth import AuthData

class Authorization:

    def __init__(self, endpoint:str, client:object, auth_data: AuthData) -> None:
        self.client = client

        self.raiseException = client.raiseException
        self.defaults = client.defaults

        self.auth_data = auth_data
        self.endpoint = endpoint
        self.proxies = client.proxies
        self.structure = client.structure
        self.scheme = client.scheme

    def getVersion(self):
        """
        Returns name and version of the responsible micro service
        """

        return Utils._getServiceVersion(self, 'authorization')

    def _resolve_where(self, where:str):
        resolvedFilter = ''
        if where != None: 
            resolvedFilter = f'({Utils._resolveWhere(self, where)["topLevel"]})'
        
        return resolvedFilter

    def roles(self, nameFilter: str = None) -> pd.DataFrame:
        """
        Returns a DataFrame of available roles.

        Parameters:
        -----------
        nameFilter : str, optional
            Filters the roles by role name.
        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with name, id and revision of the role.
        """
        key = 'roles'

        _fields = '''
            id
            name
            revision
        '''

        where_string = "" 
        if nameFilter is not None:
            where_string = f'(where:{{name:{{eq:"{nameFilter}"}}}})'

        graphQLString = f'''query roles {{
            {key} {where_string}
            {{ 
                {_fields}
            }}
        }}'''

        result = Utils._executeGraphQL(self, graphQLString)

        df = pd.json_normalize(result[key])

        return df

    def rules(
        self,
        fields:list=None, 
        where:str=None
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of available rules
        """

        key = 'rules'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                filter
                revision
                role {{
                    id
                    name
                }}
                accountReference {{
                    ...on GroupReference {{
                        group {{
                            id
                            name
                        }}
                    }}
                    ...on ServiceAccountReference {{
                        serviceAccount {{
                            id
                            name
                        }}
                    }}
                }}
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query Rules {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df
    
    def addUsers(self, provider_user_ids: list, fields: list = None) -> pd.DataFrame:
        """
        Adds a list of users from Authentik via the provider_users_id to the Authorization-Service users list and returns a data frame with user information.
        Fields defines the values that are returned. By default id, providerSubject, providerUserId and userId are returned.

        Parameters:
        -----------
        provider_user_ids : list
            A list of provider user IDs to add.
        fields: list, optional
            A list of fields to include in the returned DataFrame. If not specified, the default fields will be included.

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with user information (Default: id, providerSubject, providerUserId, userId).
        """

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                providerSubject
                providerUserId
                userId
            ''' 

        correlation_id = str(uuid4())
        with logger.contextualize(correlation_id = correlation_id):
            key = 'addUsers'
            graphQLString = f'''mutation {key} {{
                {key}(input: {{
                    providerUserIds: ["{'", "'.join(provider_user_ids)}"]
                }}) {{
                    users {{
                        {_fields}
                    }}
                }}
            }}'''
            result = Utils._executeGraphQL(self, graphQLString, correlation_id)
            if result is None:
                return
            
            df = pd.json_normalize(result[key]['users'])

            return df
        

    
    def addUsersToGroups(self, user_ids: list, group_ids: list, fields: list =  None) -> pd.DataFrame:
        """
        Adds one or more users to one or more groups and returns a pandas DataFrame with user information.
        Fields defines the values that are returned. By default userId and groupIds are returned.

        Parameters:
        -----------
        user_ids : list
            A list of user IDs to add to the groups.
        group_ids : list
            A list of group IDs to add the users to.
        fields : list, optional
            A list of fields to include in the returned DataFrame. If not specified, the default fields will be included.

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with user and group information (Default:userId, groupIds).
        """
        correlation_id = str(uuid4())

        if fields is not None:
            if not isinstance(fields, list):
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)
        else:
            _fields = '''
                userId
                groupIds
            '''

        with logger.contextualize(correlation_id=correlation_id):
            key = 'addUsersToGroups'

            user_inputs = []
            for user_id in user_ids:
                for group_id in group_ids:
                    user_inputs.append(f'{{ userId: "{user_id}", groupIds: "{group_id}" }}')
            user_inputs_str = ', '.join(user_inputs)

            graphQLString = f'''mutation {{
                {key}(input: {{users: [{user_inputs_str}], }}) {{
                    users {{
                        {_fields}
                    }}
                }}
            }}'''
            result = Utils._executeGraphQL(self, graphQLString, correlation_id)

            if result is None:
                return pd.DataFrame()

            users = result[key]['users']
            df = pd.DataFrame(users)

        return df


    def getAvailableUsers(self, usernames: list = None, fields: list = None ) -> pd.DataFrame:
        """
        Retrieves available users, including all Users in Authentik, and returns a list with user information. Users can be filtered via the 'usernames' parameter.
        Fields defines the values that are returned. By default providerUserId, eMail and username are returned.
        The username will always be returned.

        Parameters:
        -----------
        usernames : list, optional
            A list of usernames to filter the results by.
        fields : list, optional
            A list of fields to include in the returned DataFrame. If not specified, the default fields will be included.

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with user and group information (Default:providerUserId, eMail).
        """

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                providerUserId
                eMail
            ''' 

        correlation_id = str(uuid4())

        with logger.contextualize(correlation_id=correlation_id):
            key = 'availableUsers'
            graphQLString = f'''query {key} {{
                {key} {{
                    username
                    {_fields}
                }}
            }}'''
            result = Utils._executeGraphQL(self, graphQLString, correlation_id)

            if result is None:
                return []

            users = result[key]

            if usernames is not None:
                filtered_users = []
                for user in users:
                    if user['username'] in usernames:
                        filtered_users.append(user)
            users = filtered_users

            df = pd.json_normalize(users)

            return df
        
  
    def addGroups(self, group_names: list, fields: list = None) -> pd.DataFrame:
        """
        Adds a list of groups and returns a pandas DataFrame with group information.
        Fields defines the values that are returned. By default name and id are returned

        Parameters:
        -----------
        group_names : list
            A list of group names to add.
        fields : list, optional
            A list of fields to include in the results. If not specified, the default fields will be included.

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with group information (Default:name, id).
        """
        if fields is not None:
            if not isinstance(fields, list):
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)
        else:
            _fields = '''
                name
                id
            '''

        correlation_id = str(uuid4())

        with logger.contextualize(correlation_id=correlation_id):
            key = 'addGroups'
            graphQLString = f'''mutation {{
                {key}(input: {{names: ["{'", "'.join(group_names)}"] }}) {{
                    groups {{
                        {_fields}
                    }}
                }}
            }}'''
            result = Utils._executeGraphQL(self, graphQLString, correlation_id)

            if result is None:
                return pd.DataFrame()

            groups = result[key]['groups']
            df = pd.DataFrame(groups)

            return df


    def users(
        self,
        fields:list=None, 
        where:str=None) -> pd.DataFrame:
        """
        Returns a DataFrame of available users
        """

        key = 'users'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
                username
                providerName
                providerUserId
                revision
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query Users {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df


    def userGroups(self, filter: str = None) -> pd.DataFrame:
        """
        Returns a DataFrame of available user groups. User Groups can be filtered by name

        Parameters:
        -----------
        filter : str, optional
            Filters the user groups by name.

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with user group information. Returns id and name. 
        """
        key = 'groups'

        _fields = '''
            id
            name
        '''
        where_string = "" 
        if filter is not None:
            where_string = f'(where:{{name:{{eq:"{filter}"}}}})'

        graphQLString = f'''query userGroups {{
            {key}{where_string} {{
                {_fields}
            }}
        }}'''

        result = Utils._executeGraphQL(self, graphQLString)

        if result is None:
            return pd.DataFrame()

        df = pd.json_normalize(result[key])

        return df
    

    def updatePermissions(self, group: str, permissions: list, permissionType: str) -> pd.DataFrame:
        
        """
        Updates permissions of a group and returns a pandas DataFrame with permission information.

        Parameters:
        -----------
        group : str
            The name of the group to update the permissions of.
        permissions : list
            A list of permissions to update. Permission are ["ADD","DELETE","READ","UPDATE"]
        permissionType : str
            The type of permission to update. PermissionTypes are "USERS", "USERS_GROUPS", "RULES", "ROLES", "DYNAMIC_OBJECT_TYPES", "SERVICE_ACCOUNTS", "EXTERNAL_RIGHTS" and "PERMISSIONS".

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with permission information. Returns groupId, type and id.
        """
        correlation_id = str(uuid4())

       
        _fields = '''
            groupId
            type
            id
        '''

        permission_df = self.getPermissionsOfGroups(filter = group) 
        permissionId = permission_df[permission_df["permissions.type"] == permissionType]["permissions.id"].iloc[0]
        groupId = permission_df["id"].iloc[0]

        with logger.contextualize(correlation_id=correlation_id):
            key = 'updatePermissions'

            graphQLString = f'''mutation {{
                {key}(input: {{permissions: {{groupId:"{groupId}", id:"{permissionId}", permissions: {permissions}, type: {permissionType}}}, }}) {{
                    permissions {{
                        {_fields}
                    }}
                }}
            }}'''
            result = Utils._executeGraphQL(self, graphQLString, correlation_id)
            df = pd.json_normalize(result[key]['permissions'], meta =  ['groupId', 'type', 'id'])
            return df


    def addPermissions(self, group: str, permissions: list, permissionType: str) -> pd.DataFrame:
        """
        Adds one or more permissions to a group and returns a pandas DataFrame with permission information.

        Parameters:
        -----------
        group : str
            The name of the group to add the permissions to.
        permissions : list
            A list of permissions to add to the group. Permission are ["ADD","DELETE","READ","UPDATE"]
        permissionType : str
            The type of permission to add to the group. PermissionTypes are "USERS", "USERS_GROUPS", "RULES", "ROLES", "DYNAMIC_OBJECT_TYPES", "SERVICE_ACCOUNTS", "EXTERNAL_RIGHTS" and "PERMISSIONS".

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame with permission information. Returns groupId, type and id.
        """
        correlation_id = str(uuid4())

    
        _fields = '''
            groupId
            type
            id
        '''

        groupId = self.userGroups(filter = group )["id"].iloc[0]

        with logger.contextualize(correlation_id=correlation_id):
            
            key = 'addPermissions'
            graphQLString = f'''mutation {{
                {key}(input: {{permissions: {{groupId:"{groupId}", permissions: {permissions}, type: {permissionType}}}}}) {{
                    permissions {{
                        {_fields}
                    }}
                }}
            }}'''
            result = Utils._executeGraphQL(self, graphQLString, correlation_id)
            df = pd.json_normalize(result[key]['permissions'], meta =  ['groupId', 'type', 'id'])
           

        return df


    def getPermissionsOfGroups(self, filter: str = None) -> pd.DataFrame:
        """
        Retrieves all groups and their permissions, permissions can be filtered by group.

        Parameters:
        -----------
        filter : list, optional
            A list of group names to filter the results by.

        Returns:
        --------
        pd.DataFrame
            A pandas DataFrame about permissions and which user group they belong to. 
            Returns name and id of the group, as well as id, permissions, type and revision of the perimissions of each group
        """

       
        _fields =f'''
            name
            id
            permissions{{
                id
                permissions
                type
                revision
            }}
        ''' 

        correlation_id = str(uuid4())
        where_string = "" 
        if filter is not None:
            where_string = f'(where:{{name:{{eq:"{filter}"}}}})'
       
        with logger.contextualize(correlation_id=correlation_id):
            key = 'permissionsOfGroups'
            graphQLString = f'''query {key} 
            {{groups {where_string}
                {{
                    {_fields}
                }}
            }}'''
            result = Utils._executeGraphQL(self, graphQLString, correlation_id)
            df = pd.json_normalize(result["groups"], 'permissions', ["name", "id"], record_prefix="permissions.")
            return df
    
    def serviceAccounts(
        self,
        fields:list=None, 
        where:str=None
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of available service accounts.
        """

        key = 'serviceAccounts'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query ServiceAccounts {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def createRole(
        self,
        inventoryName:str, 
        roleName:str,
        userGroups:list=None, 
        objectPermissions:list=['Create', 'Delete'], 
        propertiesPermissions:list=['Read', 'Update']
        ) -> None:

        """
        Creates a role and sets all rights to all properties

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory for which the new role authorizes rights.
        roleName : str
            Name of the new role.
        userGroup : list = None
            List of user group names. If None, the role will be created without attaching user groups.
        objectPermissions : list = ['Create', 'Delete']
            Default is 'Create' and 'Delete' to allow creating and deleting items of the specified inventory.
            Other entries are not allowed.
        propertiesPermissions : list = ['Read', 'Update']
            Default is 'Read' and 'Update'. All properties will receive 
            the specified rights. Other entries are not allowed.
            Permissions are not extended on referenced inventories!
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            # Parameter validation
            try:
                self.client.structure[inventoryName]
            except:
                Utils._error(self, f"Unknown inventory '{inventoryName}'")
                return
            
            try:
                roles = self.roles()
                if roleName in list(roles['name']):
                    Utils._error(self, f"Role '{roleName}' already exists.")
                    return
            except:
                pass

            if isinstance(userGroups, str):
                userGroups = [userGroups]

            if userGroups != None:
                # 'in' is not supported therefore load all groups
                dfUserGroups = self.userGroups()
                falseUserGroups = []
                for group in userGroups:
                    if group not in list(dfUserGroups['name']):
                        falseUserGroups.append(group)
                
                if falseUserGroups:
                    Utils._error(self, f"Unknown user group(s) {falseUserGroups}")
                    return

            # Create role
            properties = self.client.structure[inventoryName]['properties']

            ppstring = '[' + ','.join(map(str.upper, propertiesPermissions)) + ']'
            props = '[\n'
            refProps = '[\n'
            for _, value in properties.items():
                if value["type"] == 'scalar':
                    props += f'{{ propertyId: {Utils._toGraphQL(value["propertyId"])}\n permissions: {ppstring} }}\n'
                elif value["type"] == 'reference':
                    refProps += f'{{ propertyId: {Utils._toGraphQL(value["propertyId"])}\n inventoryId: {Utils._toGraphQL(value["inventoryId"])}\n propertyPermissions: {ppstring}\n inventoryPermissions: [NONE]\n properties: []\n referencedProperties: []\n }}'
            props += ']'
            refProps += ']'
            
            graphQLString= f'''
            mutation AddRole($roleName: String!, $inventoryId: String!, $inventoryPermissions: [ObjectPermission!]!) {{ 
                addRoles (input: {{
                    roles: {{
                        name: $roleName
                        rootInventoryPermission: {{
                            inventoryId: $inventoryId
                            inventoryPermissions: $inventoryPermissions
                            properties: {props}
                            referencedProperties: {refProps}
                            }}
                        }}
                    }})
                    {{
                    roles {{
                        id
                    }}
                }}
            }}
            '''
            params = {
                "roleName": roleName,
                "inventoryId": self.client.structure[inventoryName]['inventoryId'],
                "inventoryPermissions": list(map(str.upper, objectPermissions)),
            }

            result = Utils._executeGraphQL(self, graphQLString, correlationId, params=params)
            if result == None: return

            # if result['addRole']['errors']:
            #     Utils._listGraphQlErrors(result, 'createInventory')
            #     return

            logger.info(f"Role {roleName} created.")

            roleId = result['addRoles']['roles'][0]['id']

            # Create rules
            if userGroups != None:
                for groupname in userGroups:
                    groupId = dfUserGroups.set_index('name').to_dict(orient='index')[groupname]['id']
                    createRuleGqlString= f'''
                    mutation AddRule($roleId: String!, $groupId: String!) {{
                        addRules (input: {{
                            rules: {{
                                roleId: $roleId
                                groupId: $groupId
                                filter: ""
                                filterFormat: EXPRESSION
                                }}
                            }})
                            {{
                            rules {{
                                ruleId
                            }}
                        }}
                    }}
                    '''
                    result = Utils._executeGraphQL(self, createRuleGqlString, correlationId, params={"roleId": roleId, "groupId": groupId})
                    if result != None:
                        logger.info(f"Rule for {roleName} and user group {group} created.")
                    else:
                        logger.error(f"Rule for {roleName} and user group {group} could not be created.")

            return

    def deleteRole(self, role:str) -> None:
        """
        Deletes a role and all related rules.
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            # Get Ids of roles and rules
            roles = self.roles().set_index('name')
            roleId = roles.loc[role,'id']

            rules = self.rules()
            rules = rules.set_index('role.name')
            try:
                ruleIds = rules.loc[role,'id']
            except:
                ruleIds = []
            if not isinstance(ruleIds, str):
                ruleIds = list(ruleIds)
            else:
                ruleIds = [ruleIds]

            # First delete rules
            if ruleIds:
                deleteRuleGraphQLString = f'''
                mutation deleteRule($ruleId: String!) {{
                    removeRule(input: {{
                        ruleId: $ruleId
                    }}) {{
                        ruleId
                    }}
                }}
                '''
                for ruleId in ruleIds:
                    result = Utils._executeGraphQL(self, deleteRuleGraphQLString, correlationId, {"ruleId": ruleId})
                    if result != None:
                        logger.info(f"Rule {ruleId} of role {role} with id {ruleId} has been deleted.")
                    else:
                        Utils._error(self, f"Rule {ruleId} of role {roleId} could not be deleted.")
                        return

            # After all rules have been deleted, delete the role
            deleteRoleGraphQLString = f'''
            mutation deleteRole($roleId: String!) {{
                removeRole(input: {{
                    roleId: $roleId
                }}) {{
                    roleId
                }}
            }}
            '''
            result = Utils._executeGraphQL(self, deleteRoleGraphQLString, correlationId, {"roleId": roleId})
            if result != None:
                logger.info(f"Role {role} with id {roleId} has been deleted.")
            else:
                Utils._error(self, f"Role {roleId} could not be deleted.")
            
            return
        

    def addRule(self, role: str, group: str, filter: str) -> str:
        """
        Adds a rule connecting role with usergroup and adds a filter to this rule.

        Parameters:
        -----------
        role : str
            The name of the role associated with the rule.
        group : str
            The name of the group to add the rule to.
        filter : str
            The filter to apply to the rule. The format must be:"Object.porpertyID=filter_value" 

        Returns:
        --------
        str
            The ID of the created rule.
        """
        roleId = self.roles(nameFilter =f'{role}')['id'].iloc[0]
        groupId = self.userGroups(filter =f'{group}')['id'].iloc[0]

        graphqlString = '''
            mutation AddRule($roleId: String!, $groupId: String!, $filter: String!) {
                addRules(input: {
                    rules: {
                        roleId: $roleId
                        groupId: $groupId
                        filter: $filter
                        filterFormat: EXPRESSION
                    }
                }) {
                    rules {
                        ruleId
                    }
                }
            }
        '''

        params = {
            "roleId": roleId,
            "groupId": groupId,
            "filter": filter
        }

        result = Utils._executeGraphQL(self, graphqlString, params=params)

        if result is not None:
            rule_id = result['addRules']['rules'][0]['ruleId']
            logger.info(f"Rule for {role} and user group {groupId} created with ID {rule_id}.")
            return rule_id
        else:
            Utils._error(self, f"Rule could not be created.")
            return ''



# -*- coding: utf-8 -*-

import logging
import warnings
from logging.handlers import QueueHandler
from logging.handlers import QueueListener
from queue import Queue
from typing import Dict
from typing import Optional
from typing import Type

from . import const
from . import emitter


class LokiQueueHandler(QueueHandler):
    """This handler automatically creates listener and `LokiHandler` to handle logs queue."""

    def __init__(self, queue: Queue, **kwargs):
        """Create new logger handler with the specified queue and kwargs for the `LokiHandler`."""
        super().__init__(queue)
        self.handler = LokiHandler(**kwargs)  # noqa: WPS110
        self.listener = QueueListener(self.queue, self.handler)
        self.listener.start()


class LokiHandler(logging.Handler):
    """
    Log handler that sends log records to Loki.

    `Loki API <https://github.com/grafana/loki/blob/master/docs/api.md>`_
    """

    emitters: Dict[str, Type[emitter.LokiEmitter]] = {
        "0": emitter.LokiEmitterV0,
        "1": emitter.LokiEmitterV1,
    }

    def __init__(
        self,
        url: str,
        tags: Optional[dict] = None,
        auth: Optional[emitter.BasicAuth] = None,
        version: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """
        Create new Loki logging handler.

        Arguments:
            url: Endpoint used to send log entries to Loki (e.g. `https://my-loki-instance/loki/api/v1/push`).
            tags: Default tags added to every log record.
            auth: Optional tuple with username and password for basic HTTP authentication.
            version: Version of Loki emitter to use.

        """
        super().__init__()

        if version is None and const.emitter_ver == "0":
            msg = (
                "Loki /api/prom/push endpoint is in the depreciation process starting from version 0.4.0.",
                "Explicitly set the emitter version to '0' if you want to use the old endpoint.",
                "Or specify '1' if you have Loki version> = 0.4.0.",
                "When the old API is removed from Loki, the handler will use the new version by default.",
            )
            warnings.warn(" ".join(msg), DeprecationWarning)

        version = version or const.emitter_ver
        if version not in self.emitters:
            raise ValueError("Unknown emitter version: {0}".format(version))
        self.emitter = self.emitters[version](url, tags, auth, token)

    def handleError(self, record):  # noqa: N802
        """Close emitter and let default handler take actions on error."""
        self.emitter.close()
        super().handleError(record)

    def emit(self, record: logging.LogRecord):
        """Send log record to Loki."""
        # noinspection PyBroadException
        try:
            self.emitter(record, self.format(record))
        except Exception:
            self.handleError(record)



import copy
from typing import Optional
from uuid import uuid4
import pandas as pd

from loguru import logger
from time import sleep
from datetime import timedelta

from .utils.ut_fileimport import FileUtils
from .utils.ut import Utils


class FileImport():

    def __init__(self, core):
        global client
        client = core
        self.raiseException = client.raiseException
  
    def importNewInventory(self, filePath:str, delimiter:str, encoding : Optional[str] = 'utf-8-sig'):        
        """
        Creates a new inventory from a CSV file
        
        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        delimiter : str
            The CSV delimiter. Choose ',', ';', or 'tab'.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.

        Example:
        --------
        >>> createInventoryFromCsv(filePath='C:\\temp\\CreateMeterData.csv', delimiter=';')          
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return
            for file in files:
                content = FileUtils._readCsvFile(file, delimiter, encoding)

                ## CHECK FILE
                if content[0][0] != 'name':
                    Utils._error(self, f"Wrong format. Expected header 'name' (for inventory) at position (0, 0).")
                    return
        
                if content[2][0] != 'name':
                    Utils._error(self, f"Wrong format. Expected header 'name' (for property) at position (2, 0).")
                    return

                inventoryName = content[1][0]

                if not inventoryName:
                    Utils._error(self, f"Inventory name missing")
                    return 
        
                ## PREPARE IMPORT
                propertyList =[]   
                boolKeys = ['nullable', 'isArray', 'isReference'] 
                keys = [item for item in content[2]]
                columns = len(keys)

                for i, row in enumerate(content):
                    if i >= 3:
                        propertyDict = {}
                        for column in range(columns):
                            if content[2][column] in boolKeys:
                                if row[column] == 'false': value = False
                                if row[column] == 'true': value = True
                            elif not row[column]: continue
                            else: value = row[column]
                            propertyDict.setdefault(content[2][column], value)
                        propertyList.append(propertyDict)

                ## IMPORT
                logger.debug(propertyList)
                result = client.createInventory(inventoryName, propertyList)
                if result == {'createInventory': {'errors': None}}: 
                    logger.info(f"Inventory {inventoryName} created.")

            return

    def importItems(
            self, 
            filePath:str, 
            inventoryName:str, 
            delimiter:str=',',
            chunkSize:int = 5000, 
            pause:int = 1, 
            encoding : Optional[str] = 'utf-8-sig'
            ) -> None:
        """
        Imports items from a CSV file. The CSV file only needs a header of
        property definitions. To connect an item to a reference item in another 
        inventory, you can provide the reference property as header and the inventoryItemId 
        or you choose a unique field from the reference inventory and use the format 
        'inventory.property' in the header. Each line below the header represents a new item.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        inventoryName : str
            The field name of the inventory.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        chunkSize : int = 5000
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading. Default is 5000 items per chunk.
        pause : int = 1
            Pause in seconds between each chunk upload to avoid overloading.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.

        Example:
        --------
        >>> importItems(filePath='C:\\temp\\Items.csv', inventoryName='meterData',
                delimiter=';')          
        """
        
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return
            for file in files:
                content = FileUtils._readCsvFile(file, delimiter, encoding)

                ## PREPARE IMPORT   
                properties = client.propertyList(inventoryName, references=True)
                logger.debug(f'Property names: {properties}')

                diff = FileUtils._comparePropertiesBasic(properties, content[0])
                if len(diff) > 0:
                    Utils._error(self, f"Unknown properties: {list(diff)}")
                    return 

                properties = client.inventoryProperties(inventoryName)

                referenceFields, referenceErrors = FileUtils._checkReferences(content[0])
                if len(referenceErrors) != 0:
                    Utils._error(f"Multilevel references such as {referenceErrors} are not supported.")
                    return
                logger.debug(f"Reference fields: {referenceFields}")

                dataType, isArray, nullable, isReference = FileUtils._analyzeProperties(inventoryName, properties)
                logger.info(f"File '{filePath}' read and properties analyzed")

                if len(referenceFields) != 0:
                    referenceMapping = FileUtils._createReferenceMapping(client, inventoryName, referenceFields, isArray, content)
                else:
                    referenceMapping = None

                items = FileUtils._createItems(content, dataType, isArray, nullable, referenceMapping)
                logger.debug(f'Basic items: {items}' )
                if len(items) == 0:
                    logger.info(f"The file '{file}' did not contain any valid items to import.")
                    continue

                # ## IMPORT
                result = client.addItems(inventoryName, items, chunkSize, pause)
                logger.info(f"{len(result)} items of file '{filePath}' imported.")

            return

    def importValidityItems(
            self, 
            filePath:str, 
            inventoryName:str, 
            delimiter:str=',',
            uniquePropertyName:str=None,
            chunkSize:int = 5000, 
            pause:int = 1, 
            encoding : Optional[str] = 'utf-8-sig'
            ) -> None:
        """
        Imports items with validity periods from a CSV file. The CSV file needs a header of
        property definitions. Specify a property that is unique to add new validity items to an
        existing parent item, alternativly use the sys_inventoryItemId. If the unique property
        is not existing yet, a new (parent) item will be created. Use the system properties 
        'sys_validFrom' and 'sys_validTo' to specify validity periods. To connect an item to a 
        reference item in another inventory, you can provide the reference property as header 
        and the inventoryItemId or you choose a unique field from the reference inventory 
        and use the format 'inventory.property' in the header. 
        Each line below the header represents a new item.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        inventoryName : str
            The field name of the inventory.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        uniquePropertyName: str = None
            If you want to add items with to an existing parent item, specify a property name that is 
            unique ord use the 'sys_inventoryItemId'.
        chunkSize : int = 5000
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading. Default is 5000 items per chunk.
        pause : int = 1
            Pause in seconds between each chunk upload to avoid overloading.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.

        Example:
        --------
        >>> importValidityItems(filePath='C:\\temp\\Items.csv', inventoryName='meterData',
        delimiter=';', uniquePropertyName='meterSerialNumber')          
        """
               
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return
            for file in files:
                content = FileUtils._readCsvFile(file, delimiter, encoding)

                ## PREPARE IMPORT   
                properties = client.propertyList(inventoryName, references=True)
                properties += ['sys_validFrom', 'sys_validTo']
                logger.debug(f'Property names: {properties}')

                diff = FileUtils._comparePropertiesBasic(properties, content[0])
                if len(diff) > 0:
                    Utils._error(self, f"Unknown properties: {list(diff)}")
                    return 

                properties = client.inventoryProperties(inventoryName)

                referenceFields, referenceErrors = FileUtils._checkReferences(content[0])
                if len(referenceErrors) != 0:
                    Utils._error(f"Multilevel references such as {referenceErrors} are not supported.")
                    return

                dataType, isArray, nullable, isReference = FileUtils._analyzeProperties(inventoryName, properties)
                logger.info(f"File '{filePath}' read and properties analyzed")

                if len(referenceFields) != 0:
                    referenceMapping = FileUtils._createReferenceMapping(client, inventoryName, 
                        referenceFields, isArray, content)
                else:
                    referenceMapping = None

                if uniquePropertyName != None:
                    parentMapping = FileUtils._createParentMapping(client, inventoryName, 
                        content, uniquePropertyName)
                else:
                    Utils._error(self, f"To import validity items, please specify a unique property")

                validityItems, newItems = FileUtils._createItems(content, dataType, isArray, nullable, 
                    referenceMapping=referenceMapping, uniqueProperty=uniquePropertyName, parentItemMapping=parentMapping)
                if len(newItems) == 0 and len(validityItems) == 0:
                    logger.info(f"The file '{file}' did not contain any valid items to import.")
                    continue
                
                # Import validity items
                if len(validityItems) > 0:
                    result = client.addValidityItemsToParents(inventoryName, validityItems, chunkSize, pause)
                    logger.info(f"{len(result)} validity items of file '{filePath}' imported.")

                # Import new items
                if len(newItems) > 0:
                    result = client.addItems(inventoryName, newItems, chunkSize, pause)
                    logger.info(f"{len(result)} items of file '{filePath}' imported.")

            return

    def importTimeSeriesItems(
        self, 
        filePath:str,  
        inventoryName:str, 
        delimiter:str=',',
        chunkSize:int = 500, 
        pause:int = 1, 
        encoding : Optional[str] = 'utf-8-sig'
        ) -> None:
        """
        Imports time series inventory items from a CSV file. The CSV file only needs a header of
        property definitions. Each line below the header represents a new time series.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        inventoryName : str
            The field name of the inventory.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        chunkSize : int = 500
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading. Default is 50 items per chunk.
        pause : int = 1
            Pause in seconds between each chunk upload to avoid overloading.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.
                 
        """

        # if timeZone == None:
        #     timeZone = core._getDefaults()['timeZone']
               
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return
            for file in files:
                content = FileUtils._readCsvFile(file, delimiter, encoding)

                ## PREPARE IMPORT
                tsProperties = ['unit', 'timeUnit', 'factor']
                for header in tsProperties:
                    if not header in content[0]:
                        Utils._error(self, f"Header {header} not found. Import aborted.")
                        return 

                properties = client.propertyList(inventoryName, references=True)
                logger.debug(f'Property names: {properties}')

                diff = FileUtils._comparePropertiesTimeSeries(properties, content[0])
                if len(diff) > 0:
                    Utils._error(self, f"Unknown properties: {list(diff)}")
                    return 

                properties = client.inventoryProperties(inventoryName)

                referenceFields, referenceErrors = FileUtils._checkReferences(content[0])

                if len(referenceErrors) != 0:
                    Utils._error(f"Multilevel references such as {referenceErrors} are not supported.")
                    return

                dataType, isArray, nullable, isReference = FileUtils._analyzeProperties(inventoryName, properties)
                logger.info(f"File '{filePath}' read and properties analyzed")

                if len(referenceFields) != 0:
                    referenceMapping = FileUtils._createReferenceMapping(client, inventoryName, referenceFields, isArray, content)
                else:
                    referenceMapping = None

                #timeSeriesItems = FileUtils._createTimeSeriesItems(content, dataType, isArray, nullable)
                timeSeriesItems = FileUtils._createItems(content, dataType, isArray, nullable, referenceMapping, type='timeSeries')
                logger.debug(f'Time series items: {timeSeriesItems}' )
                if len(timeSeriesItems) == 0:
                    logger.info(f"The file '{file}' did not contain any valid items to import.")
                    continue

                ## IMPORT
                result = client.TimeSeries.addTimeSeriesItems(inventoryName, timeSeriesItems, chunkSize, pause)
                logger.info(f"{len(result)} items of {len(timeSeriesItems)} imported.")

            return

    def importTimeSeriesData(
        self, 
        filePath:str, 
        inventoryName:str, 
        importKeyProperty:str=None, 
        delimiter:str=',', 
        timeZone:str=None, 
        dateTimeFormat:str=None, 
        fromTimepoint:str=None, 
        toTimepoint:str=None, 
        timeDelta:timedelta=None, 
        chunkSize:int=20000, 
        encoding : Optional[str] = 'utf-8-sig'
        ) -> None: 
        """
        Imports time series data from a specific CSV file or a folder with multiple
        CSV files. The first column is the timestamp index, whereas the first row
        consists of inventory item ids or an import key from the time series property
        definitions. Time series values are spanned as matrix between first column and
        first row. 

        Parameters:
        -----------
        filePath: str
            A path to a folder or a specific CSV file
        inventoryName: str
            The field name of the inventory, if not provided in the import file. 
            In the import file the inventory nameis located at position (0,0).
        importKeyProperty: str = None
            By default the inventory item id is used to find map columns with 
            values with time series. As an alternative, the content of a specific
            property can be used as header which will be mapped with the time series.
            This property should be unique.
        delimiter: str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        timeZone: str = None
            A time zone provided in IANA or isoformat (e.g. 'Europe/Berlin' or 'CET'). Defaults
            to the local time zone.
        dateTimeFormat: str = None
            Several date-time formats are supported, however, a custom format according to
            datetime.strftime() and strptime() format codes can be passed to convert the
            timestamp.
        fromTimepoint: str = None
            Specify a timestamp in isoformat from which data should be imported.
        toTimepoint: str = None
            Specify a timestamp in isoformat until which data should be imported.
        timeDelta: datetime.timedelta = None
            Define a time delta to add or substract to the original timestamp.
        chunkSize: int = 20000
            Determines the number of time series datapoints which are written per chunk. Using chunks
            can be necessary to avoid overloading.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.

        Example:
        --------
        >>> client.FileImport.importTimeSeriesData(
            filePath=file,
            inventoryName='meterData'
            importKeyProperty='name'
            delimiter=';', 
            timeZone='CET', 
            dateTimeFormat='%Y-%b-%d %H:%M:%S'
            fromTimePoint='2023-01-01'
            toTimepoint='2023-03-05'
            timeDelta=-timedelta(hours=1))
            chunkSize=20000
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            ## TIMEZONE
            timeZone = Utils._timeZone(timeZone)

            ## CHECK FILE PATH
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return

            output = {}
            
            for file in files:

                fileName = file.name
            
                output.setdefault(fileName, 
                    {'Time series': 0,
                    'Time series errors': 0,
                    'Values written': 0,
                    'Value errors': 0,
                    'Errors': None})

                content = FileUtils._readCsvFile(file, delimiter, encoding)

                tsLength = len(content) - 1

                # CONVERT DATETIME COLUMN
                if dateTimeFormat == None:
                    dateTimeFormat = FileUtils._dateFormat(content[1][0])

                for i, row in enumerate(content):
                    try:
                        if content[i][0] == '': continue
                        content[i][0] = FileUtils._convertTimestamp(content[i][0], timeZone, dateTimeFormat, timeDelta)
                    except Exception as err: 
                        if i >= 1:
                            Utils._error(self, f"Timestamp {row[0]} could not be converted. {err}")
                            return 
                        pass

                ## GET ITEM ID FROM IMPORT KEY
                if importKeyProperty != None:
                    try:
                        items = client.items(inventoryName, fields=['unit', 'resolution', 'sys_inventoryItemId', importKeyProperty])
                    except Exception as err:
                        Utils._error(self, err)
                        return 
                    names = content[0].copy()
                    del names[0]
                    items = items[items[importKeyProperty].isin(names)]
                    idMapping = {}
                    for item in items.iterrows():
                        idMapping.setdefault(item[1][importKeyProperty],item[1]['sys_inventoryItemId'])

                    if len(idMapping) == 0:
                        Utils._error(self, err)
                        return 

                    logger.debug(f"Id Mapping: {idMapping}")
                else:
                    try:
                        items = client.items(inventoryName, fields=['unit', 'resolution', 'sys_inventoryItemId'])
                        logger.debug("Inventory read for default option (import with ivnentory item ids).")
                    except Exception as err:
                        Utils._error(self, err)
                        return 

                # Get the Inventory Id
                try:                      
                    inv = client.inventories(where=f'name eq "{inventoryName}"')
                    inventoryId = inv.loc[0, 'inventoryId']
                    logger.debug(f"Found inventoryId {inventoryId} for inventory {inventoryName}.")
                except:
                    Utils._error(self, err)
                    return 

                ## VERIFY IDS, CREATE DATA_DICTS, IMPORT
                errorDict = {}
                inv = None
                tsItems = [] # only used in bulk operation

                for column in range(1, len(content[0])):
                    if importKeyProperty != None:
                        try:
                            inventoryItemId = idMapping[content[0][column]]
                        except:
                            logger.warning(f"ImportKeyProperty '{content[0][column]}' not found.")
                            output[fileName]['Time series errors'] += 1
                            errorDict.setdefault(content[0][column], 'Not found')
                            continue
                    else:
                        try:
                            inventoryItemId = content[0][column]
                        except:
                            logger.warning(f"Inventory item id {inventoryItemId} not found.")
                            errorDict.setdefault(content[0][column], 'Not found')
                            continue                

                    try:
                        properties = items[items['sys_inventoryItemId'] == inventoryItemId]
                        if properties.empty:
                            logger.warning(f"Inventory item id {inventoryItemId} not found.")
                            output[fileName]['Time series errors'] += 1
                            errorDict.setdefault(content[0][column], 'Not found')
                            continue
                    except:
                        logger.warning(f"Inventory item id {inventoryItemId} not found.")
                        continue

                    properties = properties.to_dict('records')[0]
                    timeUnit = properties['resolution'].split(' ')[-1]
                    factor = properties['resolution'].split(' ')[0]

                    tsDict = {
                            'sys_inventoryId': inventoryId,
                            'sys_inventoryItemId': None,
                            'data': {
                                'resolution': {
                                    'timeUnit': None,
                                    'factor': None
                                },
                                'unit': None,
                                'dataPoints': None
                            }
                        }

                    valueList = []
                    for i, row in enumerate(content):
                        if i >= 1:     
                
                            try:
                                if row[0] == '': continue
                                if fromTimepoint:
                                    if row[0] < fromTimepoint: continue
                                if toTimepoint:
                                    if row[0] > toTimepoint: continue
                                float(row[column])              
                                valueList.append({'timestamp': row[0], 'value': row[column]})
                            except:
                                errorDict.setdefault(content[0][column], {})
                                errorDict[content[0][column]].setdefault(row[0], row[column])
                                output[fileName]['Value errors'] += 1     

                    logger.debug(f"Value list first 5 items: {valueList[:5]}")

                    tsDict['sys_inventoryItemId'] = inventoryItemId
                    tsDict['data']['unit'] = properties['unit']
                    tsDict['data']['resolution']['timeUnit'] = timeUnit
                    tsDict['data']['resolution']['factor'] = factor  
                    tsDict['data']['dataPoints'] = valueList

                    tsItems.append(tsDict)
                    output[fileName]['Values written'] += len(valueList)

                logger.debug("Time series collection created")
                output[fileName]['Time series'] = len(tsItems)

                tsItemsEmpty = copy.deepcopy(tsItems)
                for item in tsItemsEmpty:
                    del item['data']['dataPoints']
                
                for i in range(0, tsLength, chunkSize):
                    tsChunk = copy.deepcopy(tsItemsEmpty)
                    for j, ts in enumerate(tsChunk):
                        tsChunk[j]['data'].setdefault('dataPoints', tsItems[j]['data']['dataPoints'][i : i + chunkSize])
                        
                    client.TimeSeries.setTimeSeriesDataCollection(timeSeriesData=tsChunk)    
                    logger.info(f"({int(i/chunkSize+1)}/{tsLength//chunkSize+1}) chunks imported.")            

            logger.info(f"Import finished")

            return 

    def importGroupInstanceItems(
            self, 
            filePath:str, 
            groupInventoryName:str,
            instanceInventoryName, 
            importKeyProperty, 
            delimiter:str=None, 
            chunkSize:int = 500,
            pause:int = 1, 
            encoding : Optional[str] = 'utf-8-sig'
            ) -> None:
        """
        Imports  group instance items from a CSV file. The CSV file only needs a header
        of property definitions. The first column is reserved for the groupInventoryItemId.
        Each line below the header represents a new item.

        Parameters:
        -----------
        filePath: str
            The file path of the csv file that should be imported.
        groupInventoryName: str
            The field name of the group inventory.
        instanceInventoryName: str
            The field name of the time series instance inventory belonging to the group.
        importKeyProperty: str
            Provide a property of the parent group time series that is unique.
            (The first column 'groupName' used in the CSV file.)
        delimiter: str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        chunkSize: int = 500
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading.
        pause: int = 1
            Pause in seconds between each chunk upload to avoid overloading.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.

        """
              
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return

            for file in files:
                content = FileUtils._readCsvFile(file, delimiter, encoding)

                ## PREPARE IMPORT
                properties = client.propertyList(instanceInventoryName, references=True)
                logger.debug(f'Property names: {properties}')

                diff = FileUtils._comparePropertiesBasic(properties, content[0][1:-1])
                if len(diff) > 0:
                    Utils._error(self, f"Unknown properties: {list(diff)}")
                    return 

                properties = client.inventoryProperties(instanceInventoryName)

                if importKeyProperty != None:
                    try:
                        items = client.items(groupInventoryName, fields=['sys_inventoryItemId', importKeyProperty])
                    except Exception as err:
                        Utils._error(self, err)
                        return
                
                    names = [row[0] for row in content]
                    del names[0]

                    items = items[items[importKeyProperty].isin(names)]

                    idMapping = {}
                    for item in items.iterrows():
                        idMapping.setdefault(item[1][importKeyProperty],item[1]['sys_inventoryItemId'])
                    logger.debug(f"Id mapping: {idMapping}")
                    if len(idMapping) == 0:
                        Utils._error(self, f"No item ids for importKeyProperty '{importKeyProperty}' found.")
                        return 

                dataType, isArray, nullable, isreference = FileUtils._analyzeProperties(instanceInventoryName, properties)
                logger.debug(f'Data types: {dataType}')
                logger.debug(f'Array properties: {isArray}')
                logger.debug(f'Nullable properties: {nullable}')
                logger.info(f"File '{filePath}' read and properties analyzed")

                items = FileUtils._createInstanceItems(content, dataType, isArray, nullable, idMapping)
                logger.debug(f'Instance items: {items}' )

                ## IMPORT
                result = client.TimeSeries.addTimeSeriesItemsToGroups(groupInventoryName, items, chunkSize, pause)
                logger.info(f"{len(result)} items of file '{filePath}' imported.")

            return

    def importGroupInstanceItemsWithData(
            self, 
            filePath:str, 
            groupInventoryName:str,
            instanceInventoryName:str, 
            groupKeyProperty:str=None, 
            instanceKeyProperties:list=None,
            delimiter:str=',', 
            timeZone:str=None, 
            dateTimeFormat:str=None, 
            fromTimepoint:str=None, 
            toTimepoint:str=None, 
            timeDelta:timedelta=None, 
            chunkSize=20000, 
            pause:int = 1, 
            encoding : Optional[str] = 'utf-8-sig'
            ) -> None:
        """
        Imports  group instance items from a CSV file. The CSV file only needs a header
        of property definitions. The first column is reserved for the groupInventoryItemId.
        Each line below the header represents a new item.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        groupInventoryName : str
            The field name of the group inventory.
        instanceInventoryName : str
            The field name of the time series instance inventory belonging to the group.
        groupKeyProperty : str
            Is a property of the group item to identify it, which have to be placed 
            in the first line of the file. If None, the group item id is expected. 
        instanceKeyProperties:
            One or two key properties of the instance item to identiy it uniquely. The order must be
            the same as in the import file.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        timeZone: str = None
            A time zone provided in IANA or isoformat (e.g. 'Europe/Berlin' or 'CET'). Defaults
            to the local time zone.
        dateTimeFormat: str = None
            Several date-time formats are supported, however, a custom format according to
            datetime.strftime() and strptime() format codes can be passed to convert the
            timestamp.
        fromTimepoint : str = None
            Specify a timestamp in isoformat from which data should be imported.
        toTimepoint : str = None
            Specify a timestamp in isoformat until which data should be imported.
        timeDelta: datetime.timedelta = None
            Define a time delta to add or substract to the original timestamp.
        chunkSize: int = 20000
            Determines the number of time series datapoints which are written per chunk. Using chunks
            can be necessary to avoid overloading.
        pause: int = 1
            Pause in seconds between each chunk upload to avoid overloading.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            ## TIMEZONE
            timeZone = Utils._timeZone(timeZone)

            ## CHECK FILE PATH
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return
        
            output = {}
            
            for file in files:

                fileName = file.name
            
                output.setdefault(fileName, 
                    {'Time series instances': 0,
                    'Time series instance errors': 0,
                    'Values written': 0,
                    'Value errors': 0,
                    'Errors': None})

                content = FileUtils._readCsvFile(file, delimiter, encoding)

                ## PREPARE IMPORT: Get Instance properties
                instanceProperties = []
                for i, row in enumerate(content):
                    if i == 0: continue
                    if row[0] in ['unit', 'timeUnit', 'factor']: 
                        pass
                    elif row[0] == 'data':
                        dataBegin = i + 1
                        break
                    elif row[0] == 'values':
                        dataBegin = i + 1
                        break
                    else:
                        instanceProperties.append(row[0])
                    if i > 100: 
                        Utils._error(self, f"No keyword 'data' or 'values' found")
                        return

                tsLength = len(content) - dataBegin

                # CONVERT DATETIME COLUMN
                if dateTimeFormat == None:
                    dateTimeFormat = FileUtils._dateFormat(content[dataBegin][0])

                for i, row in enumerate(content[dataBegin:]):
                    try:
                        if row[0] == '': continue
                        content[dataBegin+i][0] = FileUtils._convertTimestamp(content[dataBegin+i][0], timeZone, dateTimeFormat, timeDelta)
                    except Exception as err: 
                        if i >= 1:
                            Utils._error(self, f"Timestamp {row[0]} could not be converted. {err}")
                            return 
                        pass

                ## PREPARE IMPORT: Compare Instance properties
                properties = client.propertyList(instanceInventoryName, references=True)
                logger.debug(f'Property names: {properties}')

                diff = FileUtils._comparePropertiesBasic(properties, instanceProperties)
                if len(diff) > 0:
                    Utils._error(self, f"Unknown properties: {list(diff)}")
                    return 
                
                properties = client.inventoryProperties(instanceInventoryName)

                ## PREPARE IMPORT: Get Mapping of Import Key
                if groupKeyProperty != None:
                    try:
                        items = client.items(groupInventoryName, fields=['sys_inventoryItemId', groupKeyProperty])
                    except Exception as err:
                        Utils._error(self, err)
                        return 
                
                    names = [column for column in content[0]]
                    del names[0]

                    items = items[items[groupKeyProperty].isin(names)]

                    idMapping = {}
                    for item in items.iterrows():
                        idMapping.setdefault(item[1][groupKeyProperty],item[1]['sys_inventoryItemId'])
                    logger.debug(f"Id mapping: {idMapping}")
                    if len(idMapping) == 0:
                        Utils._error(self, f"No item ids for groupKeyProperty '{groupKeyProperty}' found.")
                        return 

                ## PREPARE IMPORT: Check dataType, Array and nullable properties
                dataType, isArray, nullable, isReference = FileUtils._analyzeProperties(instanceInventoryName, properties)
                logger.info(f"File '{filePath}' read and properties analyzed")

                itemContent = FileUtils._createInstanceItemContent(content[:dataBegin])
                tsItemContent = content[:dataBegin]
                items = FileUtils._createInstanceItems(itemContent, dataType, isArray, nullable, idMapping, transpose=True)
                tsItems = FileUtils._createInstanceItems(tsItemContent, dataType, isArray, nullable, idMapping, transpose=True)

                logger.debug(f'Instance items: {items}' )

                # Find position of instance key properties
                if instanceKeyProperties == None:
                    Utils._error(self, f"No instanceKeyProperties provided.")
                    return 

                instancePropPos = []
                for i, row in enumerate(content[:dataBegin]):
                    for property in instanceKeyProperties:
                        if property == row[0]:
                            instancePropPos.append(i)

                try:                      
                    inv = client.inventories(where=f'name eq "{instanceInventoryName}"')
                    inventoryId = inv.loc[0, 'inventoryId']
                    logger.debug(f"Found inventoryId {inventoryId} for inventory {instanceInventoryName}.")
                except:
                    Utils._error(self, f"No inventory with name '{instanceInventoryName}'.")
                    return 

                ## IMPORT: create group instance items
                errorDict = {}

                for i in range(0, len(items)):
                    try:
                        client.TimeSeries.addTimeSeriesItemsToGroups(groupInventoryName, [items[i]])
                        sleep(pause)
                    except:
                        logger.warning(f"Instance item in column {i+1} could not be created")
                        errorDict.setdefault(i+1, "Instance could not be created")
                    try:
                        if len(instanceKeyProperties) == 1:
                            x0 = content[instancePropPos[0]][i+1]
                            instanceTs = client.items(instanceInventoryName, 
                                where=f'{instanceKeyProperties[0]} eq "{x0}"')
                        if len(instanceKeyProperties) >= 2:
                            x0 = content[instancePropPos[0]][i+1]
                            x1 = content[instancePropPos[1]][i+1]
                            instanceTs = client.items(instanceInventoryName, 
                                where=f'{instanceKeyProperties[0]} eq "{x0}" and {instanceKeyProperties[1]} eq "{x1}"')
                        inventoryItemId = instanceTs.loc[0, 'sys_inventoryItemId']
                    except:
                        logger.warning(f"Instance item in column {i+1} not found")
                        output[fileName]['Time series instance errors'] += 1
                        continue
                    sleep(pause)

                    tsDict = {
                        'sys_inventoryId': inventoryId,
                        'sys_inventoryItemId': inventoryItemId ,
                        'data': {
                            'resolution': {
                                'timeUnit': tsItems[i]['timeUnit'],
                                'factor': tsItems[i]['factor']
                            },
                            'unit': tsItems[i]['unit'],
                            'dataPoints': None
                        }
                    }

                    valueList = []
                    for row in content[dataBegin:]:
                        try:
                            if fromTimepoint:
                                if row[0] < fromTimepoint: continue
                            if toTimepoint:
                                if row[0] > toTimepoint: continue
                            float(row[i+1])   
                            valueList.append({'timestamp': row[0], 'value': row[i+1]})
                        except:
                            pass
                            errorDict[i+1] = {}
                            errorDict[i+1].setdefault(row[0], row[i+1])
                            output[fileName]['Value errors'] += 1
                
                    for k in range(0, tsLength, chunkSize):     
                        tsDict['data']['dataPoints'] = valueList[k : k + chunkSize]
                        try:
                            client.TimeSeries.setTimeSeriesDataCollection(timeSeriesData=[tsDict])
                        except Exception as err:
                            logger.error(f"Time series values could not be written. Cause: {err}")
                            errorDict[i+1] = err
                            break

                    output[fileName]['Values written'] += len(valueList)
                if len(errorDict) > 0:
                    output[fileName]['Errors'] = errorDict
                            
            return output

    def importReferences(
        self, 
        filePath:str, 
        delimiter:str=',', 
        encoding : Optional[str] = 'utf-8-sig'
        ) -> dict:
        """
        Connects reference fields of existing items of two inventories. 
        Connect array-type [n:1] and non array references [1:1].
        Use FileImport.ImportReferences() to generate a csv template.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        encoding : Optional[str] = 'utf-8-sig'
            The encoding of the CSV file.    Defaults to 'utf-8-sig'.

        Example:
        --------
        >>> filePath = r'C:\\temp\\importReferences.csv'
        >>> client.FileImport.importReferences(filePath, delimiter=';')
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            def _readInventory(
                    inventoryName:str, 
                    property:str,
                    items:list) -> pd.DataFrame:

                filterList = '['
                for item in items:
                    filterList += f'"{item}",' 
                filterList += ']' 
                
                df = client.items(inventoryName, fields=[property,'sys_inventoryItemId'], 
                    where=f'{property} in {filterList}')
                df.set_index(property, inplace=True)
                logger.debug(f"Items to reference (first 5):\n {df.iloc[:5]}")
                return df

            def _mapItems(
                    content:list, 
                    refItems:pd.DataFrame, 
                    array:bool=True) -> dict:

                data = {}
                for i, row in enumerate(content):
                    if i <= 1: continue
                    if row[0] == '':
                        msg = f"No value in line {i}"
                        logger.warning(msg)
                        errors.setdefault(f'Row {i}', 'No value')
                        continue
                    try:
                        if array == True:
                            data.setdefault(row[0], [])
                            data[row[0]].append(refItems['sys_inventoryItemId'][row[1]])
                        else:
                            data.setdefault(row[0], refItems['sys_inventoryItemId'][row[1]])
                    except KeyError as err:
                        msg = f"Row {i}: unknown item {row[0]}"
                        logger.warning(msg)
                        errors.setdefault(f'Row {i}', f'unknown reference item {row[1]}')
                        continue
                    except Exception as err:
                        msg = f"Row {i}: unknown error: {row[1]}"
                        logger.warning(msg)
                        errors.setdefault(f'Row {i}', f'unknown error: {row[1]}')
                        continue
                return data

            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return

            inventories = client.inventories(fields=['name', 'inventoryId'])

            output = {}
            
            for file in files:

                fileName = file.name
                errors = {}

                output.setdefault(fileName, 
                    {'Items updated': 0,
                    'Errors': None})
                content = FileUtils._readCsvFile(file, delimiter, encoding)
                
                # Read reference field
                try:
                    inventory = content[0][0].split('.')[0]
                    referenceField = content[0][0].split('.')[1]
                except:
                    msg = f"'{content[0][0]}' is not a valid 'inventory.property' format."
                    logger.error(f"{fileName}: {msg}")
                    output[fileName]['Errors'] = msg
                    continue

                # Arrange data
                fields = [column for column in content[1]]
                logger.debug(f"Fields: {fields}")
                invFields = {}
                inventoryList = []
                for item in fields:
                    try:
                        splitItem = item.split('.')
                        invFields.setdefault(splitItem[0], splitItem[1])
                        inventoryList.append(splitItem[0])
                    except:
                        msg = f"'{item}' is not a valid 'inventory.property' format."
                        logger.error(f"{fileName}: {msg}")
                        output[fileName]['Errors'] = msg
                        continue

                # Check if inventory is in inventories
                inventoryFieldsCorrect = True
                for key in invFields.keys():
                    if key not in list(inventories['name']):
                        inventoryFieldsCorrect = False
                        msg = f"'{key}' is not a known inventory"
                        logger.error(f"{fileName}: {msg}")
                        output[fileName]['Errors'] = msg
                        continue                 
                if inventoryFieldsCorrect == False: return
                logger.debug(f"Inventories: {inventoryList}")

                # Check type of first field
                properties = client.inventoryProperties(inventory)
                properties.set_index('name', inplace=True)
                if properties['type'][referenceField] != 'reference':
                    msg = f"'{property}' is not a reference field"
                    logger.error(f"{fileName}: {msg}")
                    output[fileName]['Errors'] = msg
                    continue       

                if properties['isArray'][referenceField] == True:

                    referenceValues = [row[1] for i, row in enumerate(content) if i > 1]
                    logger.debug(f"Reference Values (first 5): {referenceValues[:5]}")

                    # Read inventory items to reference    
                    dfReferenceItems = _readInventory(inventoryList[1], invFields[inventoryList[1]], referenceValues)

                    # Create Data Dict
                    mappedItems = _mapItems(content, dfReferenceItems, array=True)
                    
                    # Read inventory items to update
                    dfUpdateItems = _readInventory(inventory, invFields[inventory], mappedItems.keys())

                    # Create connections
                    itemsUpdated = 0
                    itemsReferred = 0
                    for key, value in mappedItems.items():
                        try:
                            client.updateArrayProperty(inventory, dfUpdateItems['sys_inventoryItemId'][key], referenceField, 'insert', value)
                            if len(value) != 0:
                                itemsUpdated += 1
                            itemsReferred += len(value)
                        except Exception as err:
                            msg = f"Item: {dfUpdateItems['sys_inventoryItemId'][key]}, Value: {value}, Problem: {err}"
                            logger.warning(err)
                            errors.setdefault({dfUpdateItems['sys_inventoryItemId'][key]}, [f'Value: {value}', f'Problem: {err}'])

                    output[fileName]['Items updated'] = itemsUpdated
                    output[fileName].setdefault('Items referred', itemsReferred)
                    if len(errors) != 0:
                        output[fileName]['Errors'] = errors

                    return output

                else: #If it is not an array (1:1 reference)
                    referenceValues = [row[1] for i, row in enumerate(content) if i > 1]
                    logger.debug(f"Reference Values (first 5): {referenceValues[:5]}")

                    dfReferenceItems = _readInventory(inventoryList[1], invFields[inventoryList[1]], referenceValues)

                    # Create Data Dict
                    mappedItems = _mapItems(content, dfReferenceItems, array=False)
                        
                    # Read inventory items to update
                    dfUpdateItems = _readInventory(inventory, invFields[inventory], mappedItems.keys())

                    # Create connections
                    updateItems = []
                    for key, value in mappedItems.items():
                        item = {
                            'sys_inventoryItemId': dfUpdateItems['sys_inventoryItemId'][key],
                            referenceField: value
                        }
                        updateItems.append(item)
                    client.updateItems(inventory, updateItems)

                    output[fileName]['Items updated'] = len(updateItems)
                    if len(errors) != 0:
                        output[fileName]['Errors'] = errors

                    return output

    def updateItems(
        self,
        filePath:str,
        inventoryName,
        delimiter:str=','
        ) -> None:

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            files = FileUtils._checkFilePath(filePath, self.raiseException)
            if files == None: return

            df = pd.read_csv(filePath, sep=delimiter)

            client.updateDataFrameItems(inventoryName, df)
            return

    def templateImportReferences(
        self, 
        filePath:str=None
        ) -> pd.DataFrame:
        """Provides a template to import item references."""
        content = [
            ['updateInventory.propertyToBeUpdated', '', '// Enter inventory and property to be updated, like "marketData.qualityCriteria"'],
            ['updateInventory.uniqueProperty', 'referenceInventory.uniquePropertyOfReference', '// Enter inventory name to be updated and a unique property like "marketData.name" and a referring inventory with another unique property like "qualityCriteria.name"'],
            ['Swissgrid AE CH long', 'Swissgrid AE CH long', '// Example data, repeat entries and use same item in first column to import array items'],
            ['Swissgrid AE CH short', 'Swissgrid AE CH short','// Example data, repeat entries and use same item in first column to import array items']
        ]
        header = ['column 1', 'column 2', 'Remarks']

        df = pd.DataFrame(content, columns=header)
        if filePath != None:
            df.to_csv(filePath, index=False, header=False)
        return df
       

            


import datetime
from typing import (
    Any,
    Dict,
)
from uuid import uuid4
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Union
import pytz
import os
import base64
import math
import pandas as pd
from gql import Client, gql
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient
from tzlocal import get_localzone
from loguru import logger
from collections import namedtuple
from gql.transport.requests import RequestsHTTPTransport
from .ut_auth import AuthData


if os.name == 'nt':  # only if the platform is Windows, pyperclip will be imported
    import pyperclip


@dataclass
class Defaults():
    timeZone: str = 'UTC'
    useDateTimeOffset: bool = True
    copyGraphQLString: bool = False


class Structure():
    queryStructure = f'''query inventories {{
    inventories (pageSize:1000) {{
        name
        inventoryId
        displayValue
        isDomainUserType
        hasValidityPeriods
        historyEnabled
        propertyUniqueness {{
            uniqueKey
            properties
            }}
        variant {{
            name
            properties {{
                name
                type
                isArray
                nullable
                }}
            }}
        properties {{
            name
            ...Scalar
            type
            isArray
            nullable
            propertyId
            ... Reference 
            }}
        }}
    }}
    fragment Scalar on IScalarProperty {{
        dataType
        }}
    fragment Reference on IReferenceProperty {{
        inventoryId
        inventoryName
        }}
    '''

    def _introspectionQueryString():
        introspectionQueryString = r'''
            query IntrospectionQuery { __schema { queryType { name } mutationType 
                { name } subscriptionType { name } types { ...FullType } directives
                { name description locations args { ...InputValue } } } }

            fragment FullType on __Type { kind name description fields(includeDeprecated: true) { name description args 
                { ...InputValue } type { ...TypeRef } isDeprecated deprecationReason } inputFields { ...InputValue } interfaces 
                { ...TypeRef } enumValues(includeDeprecated: true) { name  } possibleTypes { ...TypeRef } } 

            fragment InputValue on __InputValue { name description type { ...TypeRef } defaultValue } 
            fragment TypeRef on __Type { kind name ofType { kind name ofType { kind name ofType 
                    { kind name ofType { kind name ofType { kind name ofType { kind name ofType { kind name } } } } } } } }
        '''
        return introspectionQueryString

    def _fullStructureDict(structure) -> dict:
        """Converts the query of all inventories with all fields into a pure dict"""

        def subdict(inputObject, name):
            itemDict = {}
            for item in inputObject:
                itemDict.setdefault(item[name], {})
                for k, v in item.items():
                    itemDict[item[name]].setdefault(k, v)
            return itemDict

        structureDict = {}
        for inventory in structure['inventories']:
            inventoryName = inventory['name']
            structureDict.setdefault(inventoryName, {})
            for definitionKey, definitionValue in inventory.items():
                if not isinstance(definitionValue, (list, dict)):
                    structureDict[inventoryName].setdefault(
                        definitionKey, definitionValue)
                else:
                    if definitionKey == 'properties':
                        subDict = subdict(inventory[definitionKey], 'name')
                        structureDict[inventoryName].setdefault(
                            definitionKey, subDict)
                    if definitionKey == 'propertyUniqueness':
                        subDict = subdict(
                            inventory[definitionKey], 'uniqueKey')
                        structureDict[inventoryName].setdefault(
                            definitionKey, subDict)
                    if definitionKey == 'variant':
                        structureDict[inventoryName].setdefault(
                            definitionKey, {})
                        structureDict[inventoryName][definitionKey].setdefault(
                            'name', definitionValue['name'])
                        subDict = subdict(
                            inventory[definitionKey]['properties'], 'name')
                        structureDict[inventoryName][definitionKey].setdefault(
                            'properties', subDict)
        return structureDict

    def _fullStructureNT(structure: dict) -> namedtuple:
        """
        Provides the complete data structure of dynamic objects as named tuple. 
        Needs structureDict first
        """
        def _subItem(object: dict):
            Item = namedtuple('Item', object.keys())
            itemDict = {}
            for key, value in object.items():
                if isinstance(value, dict):
                    subItem = _subItem(value)
                    itemDict.setdefault(key, subItem)
                else:
                    itemDict.setdefault(key, value)
            item = Item(**itemDict)
            return item

        Item = namedtuple('Item', structure.keys())
        itemDict = {}
        for key, value in structure.items():
            if isinstance(value, dict):
                subItem = _subItem(value)
                itemDict.setdefault(key, subItem)
            else:
                itemDict.setdefault(key, value)
        return Item(**itemDict)

    def _inventoryNT(structure) -> namedtuple:
        """
        Provides a simplified namedtuple of dynamic objects for interactive usage
        """
        inventoryDict = {key: key for key in structure.keys()}
        Inventories = namedtuple('Inventories', inventoryDict.keys())
        return Inventories(**inventoryDict)

    def _inventoryPropertyNT(structure) -> namedtuple:
        """
        Provides a simplified namedtuple of inventory properties for interactive usage
        """
        Inventory = namedtuple('Inventories', structure.keys())
        inventoryDict = {}

        for inventory in structure.keys():
            propertyDict = {}
            for key in structure[inventory]['properties'].keys():
                propertyDict.setdefault(key, key)
                Properties = namedtuple('Properties', propertyDict.keys())
                properties = Properties(**propertyDict)
            inventoryDict.setdefault(inventory, properties)
        return Inventory(**inventoryDict)


class Utils():
    errors = f'''errors {{
                    message
                    code
                }}'''

    def _error(self, msg: str):
        if self.raiseException:
            raise Exception(msg)
        else:
            logger.error(msg)
            return

    def _timeZone(timeZone):
        """
        Returns the timezone string based on the input.

        If the input is None or 'local', it attempts to get the local timezone.
        If the local timezone cannot be determined, it falls back to the default local timezone.
        Otherwise, it returns the timezone string for the given timezone.

        Args:
            timeZone (str): The timezone to be converted to a string. If None or 'local', the local timezone is used.

        Returns:
            str: The string representation of the timezone.
        """
        if timeZone == None or timeZone == 'local':
            try:
                localTimeZone = get_localzone().zone
                return str(pytz.timezone(localTimeZone))
            except AttributeError:
                return str(get_localzone())
        else:
            return str(pytz.timezone(timeZone))

    def _graphQLList(itemList: list) -> str:
        """Converts a list to a graphQL list"""

        def is_number(n):
            try:
                float(n)
            except ValueError:
                return False
            return True

        result = '['
        for item in itemList:
            if is_number(item):
                result += f'{item},'
            else:
                result += f'"{item}",'
        result += ']'
        return result

    def _queryFields(fieldList: list, arrayTypeFields: list = None, arrayPageSize: int = None, filter: dict = None, recursive=False) -> str:
        """
        Transforms a Python list of fields into graphQL String of fields
        including fields of referenced inventories        
        """

        def nestedItem(item):
            itemLength = len(item)
            nonlocal fields
            line = ''
            for i in range(itemLength - 1):
                itemStr = ".".join(item[:i+1])
                if arrayTypeFields != None and itemStr in arrayTypeFields:
                    if filter != None and item[i] in filter.keys():
                        line += f'{item[i]} (pageSize: {arrayPageSize}, where: {filter[item[i]]}) {{ '
                    else:
                        line += f'{item[i]} (pageSize: {arrayPageSize}) {{ '
                else:
                    line += f'{item[i]} {{ '
            line += f'{item[-1]} '
            for _ in range(itemLength - 1):
                line += '}'
            line += ' \n'
            fields += line

        fields = ''
        splitList = [item.split('.') for item in fieldList]
        logger.debug(f"intermediate step - splitted list: {splitList}")

        for item in splitList:
            if len(item) == 1:
                fields += f'{item[0]}  \n'
            else:
                if recursive == False:
                    if item[-1] == '_displayValue':
                        nestedItem(item)
                    if item[-1] == 'sys_inventoryItemId':
                        nestedItem(item)
                else:
                    nestedItem(item)
        return fields

    def _handleWhere(self, where, inventoryName: str = None) -> str:
        """Handles the where argument if any"""
        if where != None:
            resolvedFilter = Utils._resolveWhere(self, where, inventoryName)
            logger.debug(f"Resolved filter: {resolvedFilter}")
            if 'topLevel' in resolvedFilter.keys():
                topLevelFilter = resolvedFilter['topLevel']
                if topLevelFilter == None:
                    return
        else:
            topLevelFilter = ''
            resolvedFilter = None
        if '[]' in topLevelFilter:
            Utils._error(self, "Filter is not resolvable.")
            return ''
        return topLevelFilter, resolvedFilter

    def _toGraphQL(value) -> Union[str, int, float, bool]:
        """
        Escapes a value for usage in GraphQL. Surrounding " will be added automatically.

        Parameters
            value
                The value which should be escaped for GraphQL.

        Returns
            Escaped value. Includes surrounding " for strings.

        Raises
            ValueError
                If value is NaN.

        Examples:
        * string ABC becomes string "ABC"
        * string A"BC becomes string "A\\"BC"
        * string "ABC" becomes string "\\"ABC\\""
        * int 123 becomes 123
        * float 12.34 becomes 12.34
        * None becomes string null
        * True becomes string true
        * False becomes string false
        * [1, 2, 3] becomes [1, 2, 3]
        """
        if (value == None):
            return 'null'
        if (isinstance(value, bool)):
            return 'true' if value else 'false'
        if (isinstance(value, int)):
            return str(value)
        if (isinstance(value, float)):
            if (math.isnan(value)):
                raise ValueError('Value is NaN.')
            return str(value)
        if type(value) == list:
            escapedList = ', '.join(Utils._toGraphQL(item) for item in value)
            return f'[{escapedList}]'
        escapedValue = str(value).replace('\\', '\\\\').replace('"', '\\"')
        return f'"{escapedValue}"'

    def _resolveWhere(self, filterArg, inventoryName=None):
        """
        How does this work:
        A list of lists is created, where 'or terms' are the elements of the parent list and 
        'and terms' are elements of a child list of an or (or single) element.
        For each 'and' in a child list, the string will be closed with an extra '}'

        Lists (as string) are treated seperately, but work the same as a single or an 'and' element.
        """

        def mapOperator(operator):
            operators = {
                '==': 'eq',
                'eq': 'eq',
                'in': 'in',
                '<': 'lt',
                '>': 'gt',
                '<=': 'lte',
                '>=': 'gte',
                'lt': 'lt',
                'gt': 'gt',
                'lte': 'lte',
                'gte': 'gte',
                'contains': 'contains',
                '!=': 'neq',
                'ne': 'neq',
                'neq': 'neq',
                'not in': 'not in',
                'startsWith': 'startsWith',
                'startswith': 'startsWith',
                'endsWith': 'endsWith',
                'endswith': 'endsWith',
                '=': 'eq'
            }
            if operator in operators:
                return operators[operator]
            else:
                logger.error(f"Unknown operator '{operator}'")

        def _split(itemToSplit: str) -> list:
            "Splits a string by spaces and strips all blanks"
            splitted = itemToSplit.split(' ')
            for item in reversed(splitted):
                if item == '':
                    splitted.remove(item)

            if len(splitted) > 3:
                Utils._error(
                    self, f'''Invalid filter criteria {itemToSplit}. Did you miss to put the search string in double quotes ("")?''')
            return splitted

        def _secondLevelFilter(el: namedtuple, inventoryName: str, resultDict: dict):
            """Builds the filter part for second level"""

            subProperty = el.property.split('.')[0]
            subFilterProperty = el.property.split('.')[1]
            try:
                subInventory = self.structure[inventoryName]['properties'][subProperty]['inventoryName']
            except:
                Utils._error(
                    self, f"Unknown property '{subProperty}' in '{el.property}'.")
                return

            df = self.items(subInventory, fields=['sys_inventoryItemId'],
                            where=[f'{subFilterProperty} {el.operator} {el.searchItem}'])

            try:
                itemIds = list(df['sys_inventoryItemId'])
            except:
                logger.warning(
                    f"'{subFilterProperty} {el.operator} {el.searchItem}' does not lead to a result.")
                return ''
            if len(itemIds) == 0:
                logger.error(f"{el.property} is not a valid filter criteria.")
                return ''

            # Check, if filter compund is of array type (treated differently)
            subFilterPropertyIsArray = self.structure[inventoryName]['properties'][subProperty]['isArray']
            if subFilterPropertyIsArray:
                if len(itemIds) == 1:
                    whereString = f'{{sys_inventoryItemId: {{ eq: "{itemIds[0]}" }} }}'
                else: # if it is a list
                    itemIds = Utils._graphQLList(itemIds)
                    whereString = f'{{sys_inventoryItemId: {{ in: {itemIds} }} }}'
            else:
                if len(itemIds) == 1:
                    whereString = f'{{{subProperty}: {{sys_inventoryItemId: {{ eq: "{itemIds[0]}" }} }} }}'
                else: # if it is a list
                    itemIds = Utils._graphQLList(itemIds)
                    whereString = f'{{{subProperty}: {{sys_inventoryItemId: {{ in: {itemIds} }} }} }}'

            # subFilterPropertyIsArray = self.structure[inventoryName]['properties'][subProperty]['isArray']
            if subFilterPropertyIsArray:
                resultDict.setdefault(subProperty, whereString)
                logger.warning("Test")
                Utils._error(
                    self, f"Array properties are not yet supported in filters: '{subProperty}'")
                return resultDict.setdefault(subProperty, whereString)

            else:
                return whereString

        def createFilterPart(subElement) -> namedtuple:
            SubElement = namedtuple(
                'SubElement', ['property', 'operator', 'searchItem'])
            if '[' in subElement:  # if it is a list
                x = subElement.find('[')
                sub = subElement[:x].split(' ')
                sub_last = subElement[x:]
                if sub_last.count('"') == 0:  # if it is a real list put in by {}
                    sub_last = sub_last[1:-1].split(',')
                    sub_last = [item.lstrip() for item in sub_last]
                    sub_last = [item.replace("'", "") for item in sub_last]
                    sub_last = Utils._graphQLList(sub_last)
                el = SubElement(
                    property=sub[0], operator=sub[1], searchItem=sub_last)
                if el.property.count('.') == 1:
                    whereString = _secondLevelFilter(
                        el, inventoryName, resultDict)
                else:
                    whereString = f'{{ {el.property}: {{ {mapOperator(el.operator)}: {el.searchItem} }} }}'
                return whereString
            else:
                if subElement.count('"') == 2:  # search item is a string
                    x = subElement.find('"')
                    sItem = subElement[x:]
                    sub = _split(subElement[:x])
                # search item is not a string (but valid)
                elif subElement.count('"') == 0:
                    sub = _split(subElement)
                    sItem = sub[2]
                else:  # invalid sub elements with 1 or more than 3 double quotes
                    Utils._error(self, f"Invalid filter criteria {subElement}")

                el = SubElement(property=sub[0],
                                operator=sub[1], searchItem=sItem)
                if el.property.count('.') == 1:
                    whereString = _secondLevelFilter(
                        el, inventoryName, resultDict)
                else:
                    whereString = f'{{ {el.property}: {{ {mapOperator(el.operator)}: {el.searchItem} }} }}'
                return whereString

        def filterPart(arg):
            if type(arg) == str:
                return createFilterPart(arg)
            elif type(arg) == list:
                whereString = '{and: ['
                for i, subArg in enumerate(arg):
                    whereString += filterPart(subArg)
                    if i != len(arg) - 1:
                        whereString += ', '
                whereString += ']} '
                return whereString
            elif type(arg) == tuple:
                whereString = '{or: ['
                for i, subArg in enumerate(arg):
                    whereString += filterPart(subArg)
                    if i != len(arg) - 1:
                        whereString += ', '
                whereString += ']} '
                return whereString
            elif type(arg) == bool:
                Utils._error(self, f"Wrong syntax of filter '{filterArg}'.")
                return

        resultDict = {}

        if type(filterArg) in [list, tuple]:
            whereString = filterPart(filterArg)
        else:
            if ' or ' in filterArg:
                Utils._error(
                    self, f"'or' is not supported in string type filters.")
            filterArg = filterArg.split(' and ')
            whereString = filterPart(filterArg)
            resultDict.setdefault('topLevel', 'where: ' + whereString)

        if whereString:
            resultDict.setdefault('topLevel', 'where: ' + whereString)

        logger.debug(f"ResultDict: {resultDict}")
        return resultDict

    def _propertiesToString(properties: list) -> str:
        """ Converts a list of property dicts for many items into a string """
        if type(properties) == list:
            _properties = '[\n'
            for property in properties:
                _properties += '{\n'
                for key, value in property.items():
                    _properties += Utils._customProperties(key, value)

                _properties += '}\n'
            _properties += ']'
            return _properties
        if type(properties) == dict:
            _properties = '{\n'
            for key, value in properties.items():
                _properties += Utils._customProperties(key, value)
            _properties += '}\n'
            return _properties
        else:
            logger.error(
                f"Type of property items has to be either list ord dict.")
            return

    def _tsPropertiesToString(properties: list) -> str:
        """ Converts a list of property dicts for many items into a string """
        timeUnit, factor = 'timeUnit', 'factor'
        _properties = '[\n'
        for property in properties:
            _properties += '{\n'
            for key, value in property.items():
                if key == 'resolution':
                    try:
                        _properties += f'{key}: {{\n'
                        _properties += f'timeUnit: {value[timeUnit]}\n'
                        _properties += f'factor: {value[factor]}\n'
                        _properties += f'}}\n'
                    except KeyError:
                        logger.error(
                            "Missing 'timeUnit' and/or 'factor' for Timeseries resolution")
                        return
                else:
                    _properties += Utils._customProperties(key, value)

            _properties += '}\n'
        _properties += ']'
        return _properties

    def _addToGroupPropertiesToString(groupItemId: str, properties: list) -> str:
        """ Converts a list of property dicts for many items into a string """

        _properties = '[\n'
        for property in properties:
            _properties += f'{{sys_groupInventoryItemId: "{groupItemId}"\n'
            for key, value in property.items():
                _properties += Utils._customProperties(key, value)

            _properties += '}\n'
        _properties += ']'
        return _properties

    def _uniquenessToString(propertyUniqueness: list):
        """
        Converts a list of unique keys into a string
        """

        _uniqueKeys = '[\n'
        for item in propertyUniqueness:
            key = item['uniqueKey']
            _uniqueKeys += f'{{uniqueKey: "{key}" properties: ['
            for value in item['properties']:
                _uniqueKeys += f'"{value}",'

            _uniqueKeys += ']}\n'
        _uniqueKeys += ']'
        return _uniqueKeys

    def _customProperties(key: str, value: object) -> str:
        """
        Used internally (in Utils) as helper function
        """

        if key == 'dataType':
            return f'{key}: {value}\n'
        return  f'{key}: {Utils._toGraphQL(value)}\n'
    
    def _combineWithDot(left:str, right:str) -> str:
        """
        Combines two strings with a dot in between
        """
        if left is not None and left != '':
            return left + '.' + right
        return right

    def _properties(scheme, inventoryName:str, recursive:bool=True, sysProperties:bool=False, maxRecursionDepth:int=2) -> dict:
        """
        Creates a nested (or unnested) dict with properties and array 
        type fields for further usage out of the scheme
        """
        arrayTypeFields = []

        def _getInventoryObject(scheme, inventoryName):
            for item in scheme['__schema']['types']:
                if item != None and item['name'] == inventoryName:
                    return item['fields']

        def _createDict(inv, itemPath, dict, currentRecursionLevel=0):
            inventoryObject = _getInventoryObject(scheme, inv)
            
            for item in inventoryObject:                
                itemName = item['name']
                itemTypeName = item['type']['name']
                itemKind = item['type']['kind']
                currentItemPath = Utils._combineWithDot(itemPath, itemName)
                if sysProperties == False:
                    if itemName.startswith('sys_'):
                        if itemName == 'sys_inventoryItemId':
                            pass
                        else:
                            continue
                if itemKind == 'SCALAR':
                    dict.setdefault(itemName, itemTypeName)
                elif itemKind == 'LIST':
                    if itemName == 'sys_permissions':
                        pass
                    elif item['type']['ofType']['kind'] == 'OBJECT':
                        arrayTypeFields.append(currentItemPath)
                        if recursive == False or currentRecursionLevel > maxRecursionDepth:
                            dict.setdefault(itemName, itemTypeName)
                        else:
                            dict.setdefault(itemName, _createDict(
                                item['type']['ofType']['name'], currentItemPath, {}, currentRecursionLevel + 1))
                    else:
                        dict.setdefault(itemName, itemTypeName)
                elif itemKind == 'OBJECT':
                    if recursive == False or currentRecursionLevel > maxRecursionDepth:
                        dict.setdefault(itemName, itemTypeName)
                    else:
                        dict.setdefault(itemName, _createDict(
                            itemTypeName, currentItemPath, {}, currentRecursionLevel + 1))

            return dict

        properties = _createDict(inventoryName, "", {})
        
        propertyDict = {
            'properties': properties, 
            'arrayTypeFields': arrayTypeFields
        }

        logger.debug(f"returned property dict: {propertyDict}")
        
        return propertyDict

    def _propertyList(propertyDict: dict, recursive: bool = False) -> list:
        """Creates a flat list of properties"""

        def flatten_dict(d, path=''):
            newPath = '' if path == '' else f'{path}.'
            result = []
            for k, v in d.items():
                if isinstance(v, dict):
                    if recursive:
                        result.extend(flatten_dict(v, f'{newPath}{k}'))
                    else:
                        result.append(f'{newPath}{k}._displayValue' if '_displayValue' in v else f'{newPath}{k}.sys_inventoryItemId')
                else:
                    result.append(f'{newPath}{k}')
            return result

        propertyList = flatten_dict(propertyDict)
        logger.debug(f'returend property list: {propertyList}')
        return propertyList

    def _propertyTypes(propertyDict: dict) -> dict:
        """Uses _properties() o create a flat dictionary of properties"""

        propertyTypes = {}

        def instDict(subDict, path):
            for k, v in subDict.items():
                if isinstance(v, dict):
                    path = f'{path}.{k}'
                    instDict(subDict[k], path)
                else:
                    propertyTypes.setdefault(f'{path}.{k}', v)

        for k, v in propertyDict.items():
            if isinstance(v, dict):
                instDict(propertyDict[k], k)
            else:
                propertyTypes.setdefault(k, v)

        logger.debug(f"returned property types: {propertyTypes}")
        return propertyTypes

    def _orderItems(self, orderBy: Union[dict, list, str], asc: Union[list, str] = None) -> str:

        def _orderDictToString(orderBy: dict) -> str:

            order = 'order: ['
            for key, value in orderBy.items():
                if value not in ['ASC', 'DESC']:
                    Utils._error(
                        self, f"Invalid value '{value}' for property '{key}'. Use 'ASC' or 'DESC' instead.")
                    return
                order += f'{{{key}: {value}}}'
            order += ']'
            return order

        mapping = {False: 'DESC', True: 'ASC', None: 'ASC'}
        if orderBy != None:
            if type(orderBy) == dict:
                order = _orderDictToString(orderBy)
            if type(orderBy) == list:
                if type(asc) == bool:
                    _orderBy = {property: mapping[asc] for property in orderBy}
                elif type(asc) == list:
                    _asc = [mapping[item] for item in asc]
                    _orderBy = dict(zip(orderBy, _asc))
                elif asc == None:
                    _orderBy = {property: 'ASC' for property in orderBy}
                else:
                    Utils._error(
                        self, f"Invalid type '{asc}' for order direction. Use bool or list of bools.")
                order = _orderDictToString(_orderBy)
            if type(orderBy) == str:
                if type(asc) == bool:
                    _orderBy = {orderBy: mapping[asc]}
                elif type(asc) == list:
                    _orderBy = {orderBy: mapping[asc[0]]}
                elif asc == None:
                    _orderBy = {orderBy: 'ASC'}
                else:
                    Utils._error(
                        self, f"Invalid type '{asc}' for order direction. Use bool instead.")
                order = _orderDictToString(_orderBy)
        else:
            order = ''

        return order

    def _copyGraphQLString(graphQLString: str, copyGraphQLString: bool = False) -> None:
        """Can be applied to any core function to get the GraphQL string which is stored in the clipboard"""
        if copyGraphQLString == True and os.name == 'nt':
            return pyperclip.copy(graphQLString)

    def _getVariantId(variants: pd.DataFrame, name: str) -> str:
        """ Gets the variant Id from a given name"""
        variants.set_index('name', inplace=True)
        return variants.loc[name][0]

    def _listGraphQlErrors(result: dict, key: str) -> None:
        """Print errors from GraphQL Query to log"""

        for i in result[key]['errors']:
            logger.error(i['message'])

    def _encodeBase64(file: str):
        with open(file) as file:
            content = file.read()
            content = base64.b64encode(content.encode('ascii'))
            return content.decode('UTF8')

    # deprecated:
    def _getInventoryId(self, inventoryName):
        inventory = self.inventories(where=f'name eq "{inventoryName}"')

        if inventory.empty:
            msg = f"Unknown inventory '{inventoryName}'."
            if self.raiseException:
                raise Exception(msg)
            else:
                logger.error(msg)
                return

        inventoryId = inventory.loc[0, 'inventoryId']
        logger.debug(f'returned inventory id: {inventoryId}')
        return inventoryId

    def _arrayItemsToString(arrayItems: list, operation: str, cascadeDelete: bool) -> str:
        """Converts a list of array items to a graphQL string"""

        cDelValue = 'true' if cascadeDelete == True else 'false'

        if operation == 'insert':
            _arrayItems = 'insert: [\n'
            for item in arrayItems:
                _arrayItems += f'{{value: "{item}"}}\n'
            _arrayItems += ']'
            return _arrayItems
        if operation == 'removeByIndex':
            _arrayItems = f'cascadeDelete: {cDelValue}\n'
            _arrayItems += 'removeByIndex: ['
            for item in arrayItems:
                _arrayItems += f'{item}, '
            _arrayItems += ']'
            return _arrayItems
        if operation == 'removeById':
            _arrayItems = f'cascadeDelete: {cDelValue}\n'
            _arrayItems += 'removeById: ['
            for item in arrayItems:
                _arrayItems += f'"{item}", '
            _arrayItems += ']'
            return _arrayItems
        if operation == 'removeAll':
            _arrayItems = f'cascadeDelete: {cDelValue}\n'
            _arrayItems += 'removeByIndex: ['
            for item in arrayItems:
                _arrayItems += f'{item}, '
            _arrayItems += ']'
            return _arrayItems

    def _executeGraphQL(
            self,
            graphQLString,
            correlationId: str = None,
            params: Dict[str, Any] = None):
        """Executes GraphQl, this code is used in every main function"""

        cid = correlationId or str(uuid4())
        context_logger = logger.bind(correlationId=str(cid))

        Utils._copyGraphQLString(
            graphQLString, self.defaults.copyGraphQLString)
        context_logger.trace(f"GraphQLString: {graphQLString}")
        try:
            query = gql(graphQLString)
        except Exception as err:
            if self.raiseException:
                raise Exception(err)
            else:
                context_logger.error(err)
                return

        try:
            auth_data = self.auth_data
            if auth_data.token is None:
                raise Exception('No valid token available. Please login first.')
            # refresh token if necessary
            expires_at = datetime.fromtimestamp(auth_data.token['expires_at'])
            if expires_at - timedelta(minutes=1) < datetime.now():
                logger.debug(f'Access token expiring in less than 1 minute')
                if auth_data.service_account_secret is None:
                    extra = { 'client_id': auth_data.client_id }
                    logger.debug(f'Refreshing token')
                    self.auth_data.token = OAuth2Session(auth_data.client_id, token=auth_data.token).refresh_token(
                        auth_data.token_url, **extra)
                else:
                    # login with service account returns no refresh token so we need a new login
                    logger.debug(f'Get new token with service account')
                    app_client = BackendApplicationClient(client_id=auth_data.client_id)
                    oauth = OAuth2Session(client=app_client)
                    self.auth_data.token = oauth.fetch_token(token_url=auth_data.token_url, 
                                                            client_id=auth_data.client_id, 
                                                            username=auth_data.service_account_name, 
                                                            password=auth_data.service_account_secret)

            headers = {
                'Authorization': 'Bearer ' + self.auth_data.token['access_token'],
                'X-Correlation-Id': cid,
            }
            transport = RequestsHTTPTransport(
                url=self.endpoint, headers=headers, verify=True, proxies=self.proxies)
            with Client(transport=transport, fetch_schema_from_transport=False) as session:
                result = session.execute(query, variable_values=params)
        except Exception as err:
            if self.raiseException:
                raise Exception(err)
            else:
                context_logger.error(err)
                return

        return result

    def _create_client(self, auth_data: AuthData):
        headers = {
            'authorization': 'Bearer ' + auth_data.token['access_token']
        }
        transport = RequestsHTTPTransport(
            url=self.endpoint, headers=headers, verify=True, proxies=self.proxies)
        return Client(transport=transport, fetch_schema_from_transport=False)

    def _argNone(arg, value, enum=False) -> str:
        """
        Creates a simple string (to be embedded in graphQlString) 
        for arguments that are None by default.
        """
        if value == None:
            return ''
        else:
            if enum == True:
                return f'{arg}: {value}'
            else:
                if type(value) == str:
                    return f'{arg}: "{value}"'
                elif type(value) == float:
                    return f'{arg}: {value}'
                elif type(value) == int:
                    return f'{arg}: {value}'
                elif type(value) == bool:
                    if value == True:
                        return f'{arg}: true'
                    else:
                        return f'{arg}: false'
                else:
                    return f'{arg}: "{str(value)}"'

    def _getServiceVersion(self, service: str):
        """
        Returns name and version of the responsible micro service
        """

        key = f'{service}ServiceInfo'
        graphQLString = f'''query version {{ 
            {key} {{
                name
                informationalVersion
            }}
        }}'''
        result = Utils._executeGraphQL(self, graphQLString)

        return f'{result[key]["name"]}: {result[key]["informationalVersion"]}'



import atexit
from loguru import logger

from .ut_auth import AuthData
from ..logging_loki import LokiQueueHandler, emitter
from . import __version__
from multiprocessing import Queue


class LogUtils():
    def _init_logging(
            url: str,
            auth_data: AuthData,
            log_level: str,
            sessionId: str,
            disableOAuth: bool,
            clientId: str = "S2O.TechStack.Python"):

        emitter.LokiEmitter.level_tag = "level"
        emitter.LokiEmitter.auth_data = auth_data
        emitter.LokiEmitter.disable_oauth = disableOAuth
        handler = LokiQueueHandler(
            Queue(),
            url=url,
            tags={"client": clientId},
            version="1"
        )

        logger.add(handler, level=log_level, serialize=True,
                   backtrace=True, diagnose=True)
        logger.configure(extra={
            "version": __version__,
            "session_id": sessionId,
        })

        def _teardown_logging(handler):
            handler.listener.stop()

        atexit.register(_teardown_logging, handler)
